from django.http import HttpResponse, HttpResponseForbidden, Http404
from django.template.response import TemplateResponse
from django.views.generic import TemplateView

from oscar.core.loading import import_module
report_forms = import_module('reports.forms', ['ReportForm'])
report_utils = import_module('reports.utils', ['GeneratorRepository'])


class DashboardView(TemplateView):
    template_name = 'reports/dashboard.html'
    
    def get(self, request, *args, **kwargs):
        if 'report_type' in request.GET:
            form = report_forms.ReportForm(request.GET)
            if form.is_valid():
                generator = _get_generator(form)
                if not generator.is_available_to(request.user):
                    return HttpResponseForbidden("You do not have access to this report")
                
                response = HttpResponse(mimetype=generator.mimetype)
                response['Content-Disposition'] = 'attachment; filename=%s' % generator.filename()
                generator.generate(response)
                return response
        else:
            form = report_forms.ReportForm()
        return TemplateResponse(request, self.template_name, {'form': form})


def _get_generator(form):
    code = form.cleaned_data['report_type']

    repo = report_utils.GeneratorRepository()
    generator_cls = repo.get_generator(code)
    if not generator_cls:
        raise Http404
    return generator_cls(start_date=form.cleaned_data['start_date'], 
                         end_date=form.cleaned_data['end_date'])



import csv

from oscar.core.loading import import_module
import_module('reports.reports', ['ReportGenerator'], locals())
import_module('voucher.models', ['Voucher'], locals())


class VoucherReportGenerator(ReportGenerator):
    
    filename_template = 'voucher-performance.csv'
    code = 'vouchers'
    description = 'Voucher performance'
    
    def generate(self, response):
        writer = csv.writer(response)
        header_row = ['Voucher code',
                      'Added to a basket',
                      'Used in an order',
                      'Total discount',
                     ]
        writer.writerow(header_row)
        
        vouchers = Voucher._default_manager.all()
        for voucher in vouchers:
            row = [voucher.code, voucher.num_basket_additions, voucher.num_orders, voucher.total_discount]
            writer.writerow(row)

    def filename(self):
        return self.filename_template



from oscar.apps.dynamic_images.exceptions import ResizerConfigurationError


class BaseResponse(object):
    def __init__(self,config,mime_type,cache,start_response):
        self.config = config
        self.mime_type = mime_type
        self.cache = cache
        self.start_response = start_response
        
    def build_response(self):
        pass


class DirectResponse(BaseResponse):
    """
    Serve the file directly, can use any caching mechanism
    """
    def build_response(self):
        """
        Serves the cached image directly.
        """
        status = '200 OK'

        data = self.cache.read()
        
        response_headers = [('Content-Type', self.mime_type),
                            ('Content-Length', str(len(data)))]
        self.start_response(status, response_headers)

        return [data]
    

class NginxSendfileResponse(BaseResponse):
    """
    This can only work with a disk-based caching system since it only returns
    headers about the file rather than the file data itself
    """
    
    def build_response(self):
        if not hasattr(self.cache, 'file_info'):
            msg = "Cache doesn't implements the method 'file_info'"
            raise ResizerConfigurationError(msg)
        if not self.config.get('nginx_sendfile_path',None):
            msg = 'Must provide nginx_sendfile_path in configuration'
            raise ResizerConfigurationError(msg)

        status = '200 OK'
        
        filename, content_length = self.cache.file_info()
        
        filename = self.config['nginx_sendfile_path'] + filename
        
        response_headers = [('Content-Type', self.mime_type),
                            ('Content-Length', content_length),
                            ('X-Accel-Redirect', filename)]
        
        self.start_response(status, response_headers)
        return ['']


from decimal import Decimal as D
import datetime

from django.utils import unittest

from oscar.test.helpers import create_product
from oscar.apps.discount.models import DiscountOffer, PERCENTAGE_DISCOUNT, ABSOLUTE_DISCOUNT, FINAL_PRICE 


class PercentageDiscountOfferTest(unittest.TestCase):
    
    def setUp(self):
        self.offer = DiscountOffer(discount_type=PERCENTAGE_DISCOUNT,
                                   discount_value=D('25.00'))
    
    def test_simple_discounted_price(self):
        product = create_product(D('100.00'))
        self.assertEquals(D('75.00'), self.offer._get_discount_price(product))
        
    def test_rounded_discounted_price(self):
        product = create_product(D('99.99'))
        self.assertEquals(D('74.99'), self.offer._get_discount_price(product))

        
class AbsoluteDiscountOfferTest(unittest.TestCase):
    
    def setUp(self):
        self.offer = DiscountOffer(discount_type=ABSOLUTE_DISCOUNT,
                                   discount_value=D('25.00'))
    
    def test_simple_discounted_price(self):
        product = create_product(D('100.00'))
        self.assertEquals(D('75.00'), self.offer._get_discount_price(product))
        
    def test_discount_larger_than_price_sets_price_to_zero(self):
        product = create_product(D('20.00'))
        self.assertEquals(D('0.00'), self.offer._get_discount_price(product))
        
    
class FixedPriceDiscountOfferTest(unittest.TestCase):
    
    def setUp(self):
        self.offer = DiscountOffer(discount_type=FINAL_PRICE,
                                   discount_value=D('25.00'))
    
    def test_simple_discounted_price(self):
        product = create_product(D('100.00'))
        self.assertEquals(D('25.00'), self.offer._get_discount_price(product))
        
    def test_discounted_price_when_original_is_cheaper(self):
        product = create_product(D('20.00'))
        self.assertEquals(D('25.00'), self.offer._get_discount_price(product))    
   


import csv

from oscar.core.loading import import_module
import_module('reports.reports', ['ReportGenerator'], locals())
import_module('offer.models', ['ConditionalOffer'], locals())


class OfferReportGenerator(ReportGenerator):
    
    filename_template = 'conditional-offer-performance.csv'
    code = 'conditional-offers'
    description = 'Offer performance'
    
    def generate(self, response):
        writer = csv.writer(response)
        header_row = ['Offer',
                      'Total discount',
                     ]
        writer.writerow(header_row)
        
        for offer in ConditionalOffer._default_manager.all():
            row = [offer, offer.total_discount]
            writer.writerow(row)

    def filename(self):
        return self.filename_template


from django.conf import settings
from django.utils import unittest
from oscar.apps.dynamic_images import ImageModifier
from oscar.apps.dynamic_images.exceptions import ResizerFormatException
from oscar.apps.dynamic_images.mods import AutotrimMod, CropMod, ResizeMod
import os
import Image

TEST_IMAGE = os.path.join(os.path.dirname(__file__), 'test_fixtures/test.jpg')

class ModifierTestCase(unittest.TestCase):

    def setUp(self):
        self.base_config = settings.DYNAMIC_MEDIA_CONFIG

    def tearDown(self):
        pass

    def test_process_path_nochange(self):
        path = '/path/to/image.jpg'
        modifier = ImageModifier(path,self.base_config)
        
        filename = modifier.source_filename
        params = modifier._params
        
        self.assertEquals(filename,'/path/to/image.jpg')
        self.assertEquals(params['type'],'jpg')
        
    def test_process_path_invalid_ext(self):
        path = '/path/to/image.ext'

        with self.assertRaises(ResizerFormatException):
            ImageModifier(path,self.base_config)
            
    def test_process_path_change_ext(self):
        path = '/path/to/image.jpg.to.gif'
        modifier = ImageModifier(path,self.base_config)
        
        filename = modifier.source_filename
        params = modifier._params
        
        self.assertEquals(filename,'/path/to/image.jpg')
        self.assertEquals(params['type'],'gif')
        
    def test_process_path_with_param(self):
        path = '/path/to/image.jpg.width-300.png'
        modifier = ImageModifier(path,self.base_config)
        
        filename = modifier.source_filename
        params = modifier._params
        
        self.assertEquals(filename, '/path/to/image.jpg')
        self.assertEquals(params['type'], 'png')
        self.assertEquals(params['width'], '300')
        self.assertEquals(len(params), 2)

    def test_process_path_with_params(self):
        path = '/path/to/image.jpg.width-300_height-200.png'
        modifier = ImageModifier(path,self.base_config)
        
        filename = modifier.source_filename
        params = modifier._params

        self.assertEquals(filename, '/path/to/image.jpg')
        self.assertEquals(params['type'], 'png')
        self.assertEquals(params['width'], '300')
        self.assertEquals(params['height'], '200')
        self.assertEquals(len(params), 3)
        
    def test_process_path_with_unusual_params(self):
        path = '/path/to/image.jpg.width-300_height-200_crop-400,40,1000,1000.png'
        modifier = ImageModifier(path,self.base_config)
        
        filename = modifier.source_filename
        params = modifier._params

        self.assertEquals(filename, '/path/to/image.jpg')
        self.assertEquals(params['type'], 'png')
        self.assertEquals(params['width'], '300')
        self.assertEquals(params['height'], '200')
        self.assertEquals(params['crop'], '400,40,1000,1000')
        self.assertEquals(len(params), 4)        
        
    def test_process_path_unusual_nochange_filenames(self):
        test_filenames = (
            '/path/to/image.jpg..png',
            '/path/to/image.jpg.png',
            '/path/to/image.jpg._.png',
        )
        
        for t in test_filenames:
            modifier = ImageModifier(t, self.base_config)
            filename = modifier.source_filename
            self.assertEquals(filename, t)         
            
    def test_process_path_dots_in_path(self):
        path = '/path.to/image.jpg.width-300_height-200.png'
        modifier = ImageModifier(path, self.base_config)
        
        filename = modifier.source_filename
        
        self.assertEquals(filename, '/path.to/image.jpg')
        
        path = '/path.to/another.nice/image.jpg.width-300_height-200.png'
        modifier = ImageModifier(path, self.base_config)
        
        filename = modifier.source_filename
        
        self.assertEquals(filename, '/path.to/another.nice/image.jpg')
        
    def test_process_path_dots_in_filename(self):
        path = '/path.to/image.ext.jpg.width-300_height-200.png'
        modifier = ImageModifier(path, self.base_config)
        
        filename = modifier.source_filename
        
        self.assertEquals(filename, '/path.to/image.ext.jpg')


class ImageModTestCase(unittest.TestCase):

    def setUp(self):
        self.image = Image.open(TEST_IMAGE)

    def tearDown(self):
        pass
    
    def test_correct_image(self):
        """ Sanity check to make sure it's the right test image """
        self.assertEquals(self.image.size, (1535, 1800))
        
    def test_resize(self):
        mod = ResizeMod(self.image, dict(width=200, height=200))
        im = mod.apply()
        
        self.assertEquals(im.size, (200, 200))
        self.assertEquals(self.image.size, (1535, 1800))  # Mod shouldn't change original
        
    def test_resize_rounding(self):
        mod = ResizeMod(self.image, dict(width=200))
        im = mod.apply()        
        
        self.assertEquals(im.size, (200, 235)) # Check scaling rounds up
        
        mod = ResizeMod(self.image, dict(height=200))
        im = mod.apply()
        
        self.assertEquals(im.size, (171, 200))
        
    def test_crop(self):
        mod = CropMod(self.image, dict(crop='50,25,200,200'))
        im = mod.apply()
        
        self.assertEquals(im.size, (150, 175))
        self.assertEquals(self.image.size, (1535, 1800))  # Mod shouldn't change original


from decimal import Decimal as D

from django.utils import unittest
from django.test.client import Client

from oscar.apps.shipping.methods import FreeShipping, FixedPriceShipping
from oscar.apps.shipping.models import OrderAndItemLevelChargeMethod
from oscar.apps.basket.models import Basket
from oscar.test.helpers import create_product
from oscar.test.decorators import dataProvider

class FreeShippingTest(unittest.TestCase):
    
    def test_shipping_is_free(self):
        method = FreeShipping()
        basket = Basket()
        method.set_basket(basket)
        self.assertEquals(D('0.00'), method.basket_charge_incl_tax())
        self.assertEquals(D('0.00'), method.basket_charge_excl_tax())
        
        
class FixedPriceShippingTest(unittest.TestCase):        
    
    def test_fixed_price_shipping_charges_for_empty_basket(self):
        method = FixedPriceShipping(D('10.00'), D('10.00'))
        basket = Basket()
        method.set_basket(basket)
        self.assertEquals(D('10.00'), method.basket_charge_incl_tax())
        self.assertEquals(D('10.00'), method.basket_charge_excl_tax())
        
    def test_fixed_price_shipping_assumes_no_tax(self):
        method = FixedPriceShipping(D('10.00'))
        basket = Basket()
        method.set_basket(basket)
        self.assertEquals(D('10.00'), method.basket_charge_excl_tax())
        
    shipping_values = lambda: [('1.00',), 
                               ('5.00',), 
                               ('10.00',), 
                               ('12.00',)]    
        
    @dataProvider(shipping_values)    
    def test_different_values(self, value):
        method = FixedPriceShipping(D(value))
        basket = Basket()
        method.set_basket(basket)
        self.assertEquals(D(value), method.basket_charge_excl_tax())
        
        
class OrderAndItemLevelChargeMethodTest(unittest.TestCase):
    
    def setUp(self):
        self.method = OrderAndItemLevelChargeMethod(price_per_order=D('5.00'), price_per_item=D('1.00'))
        self.basket = Basket.objects.create()
        self.method.set_basket(self.basket)
    
    def test_order_level_charge_for_empty_basket(self):
        self.assertEquals(D('5.00'), self.method.basket_charge_incl_tax())
        
    def test_single_item_basket(self):
        p = create_product()
        self.basket.add_product(p)
        self.assertEquals(D('5.00') + D('1.00'), self.method.basket_charge_incl_tax())
        
    def test_multi_item_basket(self):
        p = create_product()
        self.basket.add_product(p, 7)
        self.assertEquals(D('5.00') + 7*D('1.00'), self.method.basket_charge_incl_tax())


class ZeroFreeShippingThresholdTest(unittest.TestCase):
    
    def setUp(self):
        self.method = OrderAndItemLevelChargeMethod(price_per_order=D('10.00'), free_shipping_threshold=D('0.00'))
        self.basket = Basket.objects.create()
        self.method.set_basket(self.basket)
    
    def test_free_shipping_with_empty_basket(self):
        self.assertEquals(D('0.00'), self.method.basket_charge_incl_tax())
        
    def test_free_shipping_with_nonempty_basket(self):
        p = create_product(D('5.00'))
        self.basket.add_product(p)
        self.assertEquals(D('0.00'), self.method.basket_charge_incl_tax())


class NonZeroFreeShippingThresholdTest(unittest.TestCase):
    
    def setUp(self):
        self.method = OrderAndItemLevelChargeMethod(price_per_order=D('10.00'), free_shipping_threshold=D('20.00'))
        self.basket = Basket.objects.create()
        self.method.set_basket(self.basket)
        
    def test_basket_below_threshold(self):
        p = create_product(D('5.00'))
        self.basket.add_product(p)
        self.assertEquals(D('10.00'), self.method.basket_charge_incl_tax())
        
    def test_basket_on_threshold(self):
        p = create_product(D('5.00'))
        self.basket.add_product(p, 4)
        self.assertEquals(D('0.00'), self.method.basket_charge_incl_tax())
        
    def test_basket_above_threshold(self):
        p = create_product(D('5.00'))
        self.basket.add_product(p, 8)
        self.assertEquals(D('0.00'), self.method.basket_charge_incl_tax())



from django.core.urlresolvers import resolve

from oscar.core.loading import import_module
import_module('shipping.repository', ['Repository'], locals())


class CheckoutSessionData(object):
    u"""Class responsible for marshalling all the checkout session data."""
    SESSION_KEY = 'checkout_data'
    
    def __init__(self, request):
        self.request = request
        if self.SESSION_KEY not in self.request.session:
            self.request.session[self.SESSION_KEY] = {}
    
    def _check_namespace(self, namespace):
        if namespace not in self.request.session[self.SESSION_KEY]:
            self.request.session[self.SESSION_KEY][namespace] = {}
          
    def _get(self, namespace, key, default=None):
        u"""Return session value or None"""
        self._check_namespace(namespace)
        if key in self.request.session[self.SESSION_KEY][namespace]:
            return self.request.session[self.SESSION_KEY][namespace][key]
        return default
            
    def _set(self, namespace, key, value):
        u"""Set session value"""
        self._check_namespace(namespace)
        self.request.session[self.SESSION_KEY][namespace][key] = value
        self.request.session.modified = True
        
    def _unset(self, namespace, key):
        u"""Unset session value"""
        self._check_namespace(namespace)
        if key in self.request.session[self.SESSION_KEY][namespace]:
            del self.request.session[self.SESSION_KEY][namespace][key]
            self.request.session.modified = True
            
    def flush(self):
        u"""Delete session key"""
        self.request.session[self.SESSION_KEY] = {}
        
    # Shipping addresses    
        
    def ship_to_user_address(self, address):
        u"""Set existing shipping address id to session and unset address fields from session"""
        self._set('shipping', 'user_address_id', address.id)
        self._unset('shipping', 'new_address_fields')
        
    def ship_to_new_address(self, address_fields):
        u"""Set new shipping address details to session and unset shipping address id"""
        self._set('shipping', 'new_address_fields', address_fields)
        self._unset('shipping', 'user_address_id')
        
    def new_shipping_address_fields(self):
        u"""Get shipping address fields from session"""
        return self._get('shipping', 'new_address_fields')
        
    def user_address_id(self):
        u"""Get user address id from session"""
        return self._get('shipping', 'user_address_id')
    
    # Shipping methods
    
    def use_free_shipping(self):
        u"""Set "free shipping" code to session"""
        self._set('shipping', 'method_code', '__free__')
    
    def use_shipping_method(self, code):
        u"""Set shipping method code to session"""
        self._set('shipping', 'method_code', code)
        
    def shipping_method(self):
        u"""
        Returns the shipping method model based on the 
        data stored in the session.
        """
        code = self._get('shipping', 'method_code')
        if not code:
            return None
        return Repository().find_by_code(code)
    
    # Billing address fields
    
    def bill_to_new_address(self, address_fields):
        """
        Store address fields for a billing address.
        """
        self._set('billing', 'new_address_fields', address_fields)
    
    def new_billing_address_fields(self):
        """
        Return fields for a billing address
        """
        return self._get('billing', 'new_address_fields')
    
    def billing_address_same_as_shipping(self):
        """
        Record fact that the billing address is to be the same as
        the shipping address.
        """
        self._set('payment', 'billing_address_same_as_shipping', True)
        
    def is_billing_address_same_as_shipping(self):
        return self._get('payment', 'billing_address_same_as_shipping', False)
    
    # Payment methods
    
    def pay_by(self, method):
        self._set('payment', 'method', method)
        
    def payment_method(self):
        return self._get('payment', 'method')
    
    # Submission methods
    
    def set_order_number(self, order_number):
        self._set('submission', 'order_number', order_number)
        
    def get_order_number(self):
        return self._get('submission', 'order_number')    
    
    def set_submitted_basket(self, basket):
        self._set('submission', 'basket_id', basket.id)
        
    def get_submitted_basket_id(self):
        return self._get('submission', 'basket_id')
        



from django.contrib import admin

from oscar.core.loading import import_module
models = import_module('partner.models', ['Partner', 'StockRecord'])

class StockRecordAdmin(admin.ModelAdmin):
    list_display = ('product', 'partner', 'partner_sku', 'price_excl_tax', 'cost_price', 'num_in_stock')
    list_filter = ('partner',)
    
admin.site.register(models.Partner)
admin.site.register(models.StockRecord, StockRecordAdmin)



from django.contrib import admin
from django.db.models import get_model

class ProductRecordAdmin(admin.ModelAdmin):
    list_display = ('product', 'num_views', 'num_basket_additions', 'num_purchases')
    
class UserProductViewAdmin(admin.ModelAdmin):
    list_display = ('user', 'product', 'date_created')

class UserRecordAdmin(admin.ModelAdmin):
    list_display = ('user', 'num_product_views', 'num_basket_additions', 'num_orders', 'total_spent', 'date_last_order')

admin.site.register(get_model('analytics', 'productrecord'), ProductRecordAdmin)
admin.site.register(get_model('analytics', 'userrecord'), UserRecordAdmin)
admin.site.register(get_model('analytics', 'usersearch'))
admin.site.register(get_model('analytics', 'userproductview'), UserProductViewAdmin)



from django.conf.urls.defaults import patterns, url
from django.contrib.auth.decorators import login_required

from oscar.apps.customer.views import AccountSummaryView, OrderHistoryView, \
    OrderDetailView, OrderLineView, AddressListView, AddressCreateView, \
    AddressUpdateView, AddressDeleteView, EmailHistoryView, EmailDetailView, \
    AccountAuthView
from oscar.core.application import Application

class CustomerApplication(Application):
    name = 'customer'
    summary_view = AccountSummaryView
    order_history_view = OrderHistoryView
    order_detail_view = OrderDetailView
    order_line_view = OrderLineView
    address_list_view = AddressListView
    address_create_view = AddressCreateView
    address_update_view = AddressUpdateView
    address_delete_view = AddressDeleteView
    email_list_view = EmailHistoryView
    email_detail_view = EmailDetailView
    login_view = AccountAuthView

    def get_urls(self):
        urlpatterns = patterns('django.contrib.auth.views',
            url(r'^logout/$', 'logout', name='logout', kwargs={'next_page': '/'}),
        )

        urlpatterns += patterns('',
            url(r'^$', login_required(self.summary_view.as_view()), name='summary'),
            url(r'^login/$', self.login_view.as_view(), name='login'),
            url(r'^orders/$', login_required(self.order_history_view.as_view()), name='order-list'),
            url(r'^orders/(?P<order_number>[\w-]*)/$', login_required(self.order_detail_view.as_view()), name='order'),
            url(r'^orders/(?P<order_number>[\w-]*)/(?P<line_id>\d+)$', login_required(self.order_line_view.as_view()), name='order-line'),
            url(r'^addresses/$', login_required(self.address_list_view.as_view()), name='address-list'),
            url(r'^addresses/add/$', login_required(self.address_create_view.as_view()), name='address-create'),
            url(r'^addresses/(?P<pk>\d+)/$', login_required(self.address_update_view.as_view()), name='address-detail'),
            url(r'^addresses/(?P<pk>\d+)/delete/$', login_required(self.address_delete_view.as_view()), name='address-delete'),
            url(r'^emails/$', login_required(self.email_list_view.as_view()), name='email-list'),
            url(r'^emails/(?P<email_id>\d+)/$', login_required(self.email_detail_view.as_view()), name='email-detail'),
            )
        return urlpatterns

application = CustomerApplication()



from django.contrib import admin

from oscar.core.loading import import_module
import_module('voucher.models', ['Voucher', 'VoucherApplication'], locals())

    
class VoucherAdmin(admin.ModelAdmin):
    list_display = ('name', 'code', 'usage', 'num_basket_additions', 'num_orders', 'total_discount')    
    readonly_fields = ('num_basket_additions', 'num_orders', 'total_discount')
    fieldsets = (
        (None, {
            'fields': ('name', 'code', 'usage', 'start_date', 'end_date')
        }),
        ('Benefit', {
            'fields': ('offers',)
        }),
        ('Usage', {
            'fields': ('num_basket_additions', 'num_orders', 'total_discount')
        }),
        
    )
    
class VoucherApplicationAdmin(admin.ModelAdmin):
    list_display = ('voucher', 'user', 'order', 'date_created')
    readonly_fields = ('voucher', 'user', 'order')        


admin.site.register(Voucher, VoucherAdmin)
admin.site.register(VoucherApplication, VoucherApplicationAdmin)



from django import template
from django.db.models import get_model

register = template.Library()
Category = get_model('catalogue','category')

@register.tag(name="category_tree")
def do_category_list(parse, token):
    tokens = token.split_contents()
    error_msg = "%r tag uses the following syntax: {%% category_tree [depth=n] as categories %%}" % tokens[0]
    depth_var = '1'

    if len(tokens) == 4:
        if tokens[2] != 'as':
            raise template.TemplateSyntaxError(error_msg)
        depth_var = tokens[1].split('=')[1]
        as_var = tokens[3]
    elif len(tokens) == 3:
        if tokens[1] != 'as':
            raise template.TemplateSyntaxError(error_msg)
        as_var = tokens[2]
    else:
        raise template.TemplateSyntaxError(error_msg)
    
    return CategoryTreeNode(depth_var, as_var)


class CategoryTreeNode(template.Node):
    def __init__(self, depth_var, as_var):
        self.depth_var = template.Variable(depth_var)
        self.as_var = as_var
        
    def _build_tree(self, root, parent, data, depth):
        for category in data[depth]:
            if not parent or category.is_child_of(parent):
                node = (category, [])
                if depth < len(data) - 1:
                    self._build_tree(node[1], category, data, depth+1)
                root.append(node)
        return root
        
    def render(self, context):
        depth = int(self.depth_var.resolve(context))
        categories = Category.objects.filter(depth__lte=depth)
        
        category_buckets = [[] for i in range(depth)]
        
        for c in categories:
            category_buckets[c.depth - 1].append(c)
        category_subtree = []
        
        self._build_tree(category_subtree, None, category_buckets, 0)
        
        context[self.as_var] = category_subtree
        return ''



from oscar.apps.dynamic_images.cache import DiskCache
from oscar.apps.dynamic_images.exceptions import ResizerConfigurationError, \
    ResizerSyntaxError, ResizerFormatError
from oscar.apps.dynamic_images.mods import AutotrimMod, CropMod, ResizeMod
from oscar.apps.dynamic_images.response_backends import DirectResponse
from wsgiref.util import request_uri, application_uri
import Image
import cStringIO
import datetime
import math
import os
import sys
import re

try:
    import cStringIO as StringIO
except:
    import StringIO

def get_class(kls):
    try:
        parts = kls.split('.')
        module = ".".join(parts[:-1])
        m = __import__(module)
        for comp in parts[1:]:
            m = getattr(m, comp)
        return m
    except (ImportError, AttributeError), e:
        raise ResizerConfigurationError('Error importing class "%s"' % kls)


def error404(path, start_response):
    """ Returns an error 404 with text giving the requested URL. """
    status = '404 NOT FOUND'
    output = '404: File Not Found: ' + path + '\n'

    response_headers = [('Content-type', 'text/plain')]
    start_response(status, response_headers)

    return [output]


def error500(path, e, start_response):
    """ Returns an error 500 with text giving the requested URL. """
    status = '500 Exception'
    output = '500: ' + str(e) + '\n'

    response_headers = [('Content-type', 'text/plain')]
    start_response(status, response_headers)

    return [output]


class ImageModifier(object):
    """
    Modifies an image and saves to a cache location

    Output Formats:

    extension => ('format', 'mime-type')

    the key is the extension appended to the URL,
    the value tuple holds PIL's name for the format and the mime-type to serve
    with
    """
    output_formats = {
        'jpeg': ('JPEG', 'image/jpeg'),
        'jpg': ('JPEG', 'image/jpeg'),
        'gif': ('GIF', 'image/gif'),
        'png': ('PNG', 'image/png'),
    }

    # When we process an image, these modifications are applied in order
    installed_modifications = (
        AutotrimMod,
        CropMod,
        ResizeMod,
    )

    quality = 80
    
    process_check = re.compile(r'^(?P<filename>.+?)\.(?P<params>[a-z0-9]+-[a-z0-9]+(.+?)*)\.(?P<type>[a-z0-9]+)$').search
    convert_check = re.compile(r'^(?P<filename>.+?)\.to\.(?P<type>[a-z0-9]+)$').search

    def __init__(self, url, config):
        if config.get('installed_mods'):
            self.installed_modifications = config['installed_mods']

        self._url = url
        self._image_root = config['asset_root']
        self._process_path()

    def _process_path(self):
        """
        Extracts parameters from the image path

        Valid syntax:
        - /path/to/image.ext (serve image unchanged)
        - /path/to/image.ext.to.newext (change format of image)
        - /path/to/image.ext.options-string.newext (change format and modify
          image)

        Format of options string is:
        key1-value1_key2-value2_key3-value3
        """
        path, name = os.path.split(self._url)
        
        p_result = self.process_check(name)
        c_result = self.convert_check(name)
        
        if p_result:
            filename = p_result.group('filename')
            type = p_result.group('type')            
            params = p_result.group('params')
            
            param_parts = params.split('_')
            
            try:
                params = dict(
                    [(x.split("-")[0], x.split("-")[1]) for x in param_parts])
            except IndexError:
                raise ResizerSyntaxError("Invalid filename syntax")            
        elif c_result:
            filename = c_result.group('filename')            
            type = c_result.group('type')
            params = {}
        else:
            filename = self._url
            type = os.path.splitext(name)[1][1:]

            params = {}
        
        params['type'] = type
        
        self._params = params
        self.source_filename = os.path.join(path,filename)

        if self._params['type'] not in self.output_formats:
            raise ResizerFormatError("Invalid output format")

    def source_path(self):
        return os.path.join(self._image_root, self.source_filename)

    def generate_image(self):
        source = Image.open(self.source_path())

        if (self._params['type'] == 'png'):
            source = source.convert("RGBA")
        else:
            source = source.convert("RGB")

        # Iterate over the installed modifications and apply them to the image
        for mod in self.installed_modifications:
            source = mod(source, self._params).apply()

        output = StringIO.StringIO()

        source.save(output, self.get_type()[0], quality=self.quality)

        output.seek(0)
        data = output.read()

        return data

    def get_type(self):
        return self.output_formats[self._params['type']]
    
    def get_mime_type(self):
        return self.get_type()[1]


class BaseImageHandler(object):
    """
    This can be called by a WSGI script, or via DjangoImageHandler.
    Django version is handy for local development, but adds unnecessary
    overhead to production
    """
    modifier = ImageModifier
    cache = DiskCache
    response_backend = DirectResponse
    
    def build_sendfile_response(self, metadata, modifier, start_response):
        pass

    def __call__(self, environ, start_response):
        path = environ.get('PATH_INFO')
        config = environ.get('MEDIA_CONFIG')

        # Don't bother processing stuff we know is invalid
        if path == '/' or path == '/favicon.ico':
            return error404(path, start_response)
        path = path[1:]

        if config.get('cache_backend'):
            self.cache = get_class(config['cache_backend'])
        if config.get('response_backend'):
            self.response_backend = get_class(config['response_backend'])
        if config.get('installed_mods'):
            mod_overrides = []
            for v in config['installed_mods']:
                mod_overrides.append(get_class(v))
            config.installed_mods = tuple(mod_overrides)            
            
        try:
            c = self.cache(path, config)
            m = self.modifier(path, config)
        except (ResizerSyntaxError, ResizerFormatError), e:
            return error500(path, e, start_response)
        try:
            if not c.check(m.source_path()):
                data = m.generate_image()
                c.write(data)
            return self.response_backend(config, m.get_mime_type(),c,start_response).build_response()
        except Exception, e:
            return error404(path, start_response)


class DjangoImageHandler(BaseImageHandler):

    def __call__(self, request):
        from django.http import HttpResponse
        from django.conf import settings

        env = request.META.copy()

        config = settings.DYNAMIC_MEDIA_CONFIG
        prefix = settings.DYNAMIC_URL_PREFIX

        env['MEDIA_CONFIG'] = config
        env['PATH_INFO'] = env['PATH_INFO'][len(prefix) + 1:]

        django_response = HttpResponse()

        def start_response(status, headers):
            status = status.split(' ', 1)[0]
            django_response.status_code = int(status)
            for header, value in headers:
                django_response[header] = value

        response = super(DjangoImageHandler, self).__call__(env, start_response)
        django_response.content = "\n".join(response)

        return django_response



from decimal import Decimal

from django.db import models
from django.utils.translation import ugettext_lazy as _
from django.template.defaultfilters import slugify
from django.conf import settings

from oscar.apps.shipping.methods import ShippingMethod


class AbstractOrderAndItemLevelChargeMethod(models.Model, ShippingMethod):
    u"""
    Standard shipping method
    
    This method has two components: 
    * a charge per order
    * a charge per item
    
    Many sites use shipping logic which fits into this system.  However, for more
    complex shipping logic, a custom shipping method object will need to be provided
    that subclasses ShippingMethod.
    """
    code = models.CharField(max_length=128, unique=True)
    name = models.CharField(_("Name"), max_length=128)
    description = models.TextField(_("Description"), blank=True)
    price_currency = models.CharField(max_length=12, default=settings.OSCAR_DEFAULT_CURRENCY)
    price_per_order = models.DecimalField(decimal_places=2, max_digits=12, default=Decimal('0.00'))
    price_per_item = models.DecimalField(decimal_places=2, max_digits=12, default=Decimal('0.00'))
    
    # If basket value is above this threshold, then shipping is free
    free_shipping_threshold = models.DecimalField(decimal_places=2, max_digits=12, blank=True, null=True)
    
    _basket = None
    
    def save(self, *args, **kwargs):
        if not self.code:
            self.code = slugify(self.name)
        super(AbstractOrderAndItemLevelChargeMethod, self).save(*args, **kwargs)
    
    class Meta:
        abstract = True
        
    def __unicode__(self):
        return self.name
    
    def set_basket(self, basket):
        self._basket = basket
    
    def basket_charge_incl_tax(self):
        u"""
        Return basket total including tax
        """
        if self.free_shipping_threshold != None and self._basket.total_incl_tax >= self.free_shipping_threshold:
            return Decimal('0.00')
        
        charge = self.price_per_order
        for line in self._basket.lines.all():
            charge += line.quantity * self.price_per_item
        return charge
    
    def basket_charge_excl_tax(self):
        u"""
        Return basket total excluding tax.  
        
        Default implementation assumes shipping is tax free.
        """
        return self.basket_charge_incl_tax()


import logging
import sys
from optparse import make_option

from django.core.management.base import BaseCommand, CommandError

from oscar.core.loading import import_module
import_module('partner.utils', ['CatalogueImporter'], locals())
import_module('partner.exceptions', ['CatalogueImportError'], locals())


class Command(BaseCommand):
    
    args = '/path/to/file1.csv /path/to/file2.csv ...'
    help = 'For creating product catalogues based on a CSV file'
    
    option_list = BaseCommand.option_list + (
        make_option('--flush',
            action='store_true',
            dest='flush',
            default=False,
            help='Flush tables before importing'),
        make_option('--delimiter',
            dest='delimiter',
            default=",",
            help='Delimiter used within CSV file(s)'),
        )

    def handle(self, *args, **options):
        logger = self._get_logger()
        if not args:
            raise CommandError("Please select a CSV file to import")
        
        logger.info("Starting catalogue import")
        importer = CatalogueImporter(logger, delimiter=options.get('delimiter'), flush=options.get('flush'))
        for file_path in args:
            logger.info(" - Importing records from '%s'" % file_path)
            try:
                importer.handle(file_path)
            except CatalogueImportError, e:
                raise CommandError(str(e))
            
    def _get_logger(self):
        logger = logging.getLogger('oscar.apps.catalogue_import')
        stream = logging.StreamHandler(self.stdout)
        logger.addHandler(stream)
        logger.setLevel(logging.DEBUG)
        return logger
        
        
        
            


import unittest

from django.test import TestCase

from oscar.core.loading import import_module, AppNotFoundError

class ImportAppTests(unittest.TestCase):

    def test_a_specified_class_is_imported_correctly(self):
        module = import_module('catalogue.models', ['Product'])
        self.assertEqual('oscar.apps.catalogue.models', module.__name__)
        
    def test_unknown_apps_raise_exception(self):
        self.assertRaises(AppNotFoundError, import_module, 'banana', ['skin'])




import logging
import sys
import os
from optparse import make_option

from django.core.management.base import BaseCommand, CommandError

from oscar.core.loading import import_module
import_module('partner.utils', ['StockImporter'], locals())
import_module('partner.exceptions', ['ImportError'], locals())


class Command(BaseCommand):
    
    args = '<partner> /path/to/file1.csv'
    help = 'For updating stock for a partner based on a CSV file'
    
    option_list = BaseCommand.option_list + (
        make_option('--delimiter',
            dest='delimiter',
            default=",",
            help='Delimiter used within CSV file(s)'),
        )

    def handle(self, *args, **options):
        if len(args) != 2:
            raise CommandError('Command requires a partner and a path to a csv file') 
        
        logger = self._get_logger()
        
        try:
            importer = StockImporter(logger, partner=args[0], delimiter=options.get('delimiter'))
            logger.info("Starting stock import")
            logger.info(" - Importing records from '%s'" % args[1])
            importer.handle(args[1])
        except ImportError, e:
            raise CommandError(str(e))
            
    def _get_logger(self):
        logger = logging.getLogger('oscar.apps.partner.import_stock')
        stream = logging.StreamHandler(self.stdout)
        logger.addHandler(stream)
        logger.setLevel(logging.DEBUG)
        return logger


from django.conf.urls.defaults import patterns, url
from django.contrib.admin.views.decorators import staff_member_required

from oscar.core.application import Application
from oscar.apps.order_management.views import OrderListView, OrderDetailView


class OrderManagementApplication(Application):
    name = 'order-management'
    
    list_view = OrderListView
    detail_view = OrderDetailView

    def get_urls(self):
        urlpatterns = patterns('oscar.order_management.views',
            url(r'^$', staff_member_required(self.list_view.as_view()), name='list'),
            url(r'^order/(?P<order_number>[\w-]*)/$', staff_member_required(self.detail_view.as_view()), name='detail'),
        )
        return urlpatterns

application = OrderManagementApplication()


from django.conf.urls.defaults import patterns, url
from django.contrib.admin.views.decorators import staff_member_required
from haystack.query import SearchQuerySet

from oscar.core.application import Application
from oscar.apps.search.views import SuggestionsView, MultiFacetedSearchView
from oscar.apps.search.search_indexes import ProductIndex
from oscar.apps.search.forms import MultiFacetedSearchForm


class SearchApplication(Application):
    name = 'search'
    
    suggestions_view = SuggestionsView
    search_view = MultiFacetedSearchView

    def get_urls(self):
        sqs = SearchQuerySet()
        for field_name, field in ProductIndex.fields.items():
            if field.faceted is True:
                # Ensure we facet the results set by the defined facetable fields
                sqs.facet(field_name)
        
        urlpatterns = patterns('',
            url(r'^suggest/$', self.suggestions_view.as_view(), name='suggest'),
            url(r'^$', self.search_view(form_class=MultiFacetedSearchForm, 
                                        searchqueryset=sqs), name='search'),
        )
        return urlpatterns

application = SearchApplication()


import datetime
from xml.dom.minidom import Document, parseString
import httplib
import urllib

from django.conf import settings
from django.db import transaction
from django.utils.translation import ugettext_lazy as _
from django.core.mail import mail_admins

from oscar.core.loading import import_module
import_module('payment.datacash.models', ['OrderTransaction'], locals())
import_module('payment.exceptions', ['TransactionDeclined', 'GatewayError', 
                                     'InvalidGatewayRequestError'], locals())

# Status codes
ACCEPTED, DECLINED, INVALID_CREDENTIALS = '1', '7', '10'


class Gateway(object):

    def __init__(self, client, password, host, cv2avs=False):
        self._client = client
        self._password = password
        self._host = host
        
        # Fraud settings
        self._cv2avs = cv2avs

    def do_request(self, request_xml):
        # Need to fill in HTTP request here
        conn = httplib.HTTPSConnection(self._host, 443, timeout=30)
        headers = {"Content-type": "application/xml",
                   "Accept": ""}
        conn.request("POST", "/Transaction", request_xml, headers)
        response = conn.getresponse()
        response_xml = response.read()
        if response.status != httplib.OK:
            raise GatewayError("Unable to communicate with payment gateway (code: %s, response: %s)" % (response.status, response_xml))
        conn.close()
        
        # Save response XML
        self._last_response_xml = response_xml
        return response_xml

    def _build_request_xml(self, method_name, **kwargs):
        """
        Builds the XML for a 'initial' transaction
        """
        doc = Document()
        req = self._create_element(doc, doc, 'Request')
        
        # Authentication
        auth = self._create_element(doc, req, 'Authentication')
        self._create_element(doc, auth, 'client', self._client)
        self._create_element(doc, auth, 'password', self._password)
            
        # Transaction    
        txn = self._create_element(doc, req, 'Transaction') 
        
        # CardTxn
        if 'card_number' in kwargs:
            card_txn = self._create_element(doc, txn, 'CardTxn')
            self._create_element(doc, card_txn, 'method', method_name)
            
            card = self._create_element(doc, card_txn, 'Card')
            self._create_element(doc, card, 'pan', kwargs['card_number'])
            self._create_element(doc, card, 'expirydate', kwargs['expiry_date'])
            
            if 'start_date' in kwargs:
                self._create_element(doc, card, 'startdate', kwargs['start_date'])
            
            if 'issue_number' in kwargs:
                self._create_element(doc, card, 'issuenumber', kwargs['issue_number'])
          
            if 'auth_code' in kwargs:
                self._create_element(doc, card, 'authcode', kwargs['auth_code'])
                
            if self._cv2avs:
                self._add_cv2avs_elements(doc, card, kwargs) 
        
        # HistoricTxn
        if 'txn_reference' in kwargs:
            historic_txn = self._create_element(doc, txn, 'HistoricTxn')
            self._create_element(doc, historic_txn, 'reference', kwargs['txn_reference'])
            self._create_element(doc, historic_txn, 'method', method_name)
            if 'auth_code' in kwargs:
                self._create_element(doc, historic_txn, 'authcode', kwargs['auth_code'])
        
        # TxnDetails
        if 'amount' in kwargs:
            txn_details = self._create_element(doc, txn, 'TxnDetails')
            if 'merchant_reference' in kwargs:
                self._create_element(doc, txn_details, 'merchantreference', kwargs['merchant_reference'])
            self._create_element(doc, txn_details, 'amount', str(kwargs['amount']), {'currency': kwargs['currency']})
        
        # Save XML for later retrieval
        self._last_request_xml = doc.toxml()
        
        return self.do_request(doc.toxml())

    def _add_cv2avs_elements(self, doc, card, kwargs):
        cv2avs = self._create_element(doc, card, 'Cv2Avs')
        if 'ccv' in kwargs:
            self._create_element(doc, cv2avs, 'cv2', kwargs['ccv'])

    def _create_element(self, doc, parent, tag, value=None, attributes=None):
        """
        Creates an XML element
        """
        ele = doc.createElement(tag)
        parent.appendChild(ele)
        if value:
            text = doc.createTextNode(value)
            ele.appendChild(text)
        if attributes:
            [ele.setAttribute(k, v) for k,v in attributes.items()]
        return ele
    
    def _get_element_text(self, doc, tag):
        try:
            ele = doc.getElementsByTagName(tag)[0]
        except IndexError:
            return None
        return ele.firstChild.data

    def _build_response_dict(self, response_xml, extra_elements=None):
        doc = parseString(response_xml)
        response = {'status': int(self._get_element_text(doc, 'status')),
                    'datacash_reference': self._get_element_text(doc, 'datacash_reference'),
                    'merchant_reference': self._get_element_text(doc, 'merchantreference'),
                    'reason': self._get_element_text(doc, 'reason')}
        if extra_elements:
            for tag, key in extra_elements.items():
                response[key] = self._get_element_text(doc, tag)
        return response

    # API

    def auth(self, **kwargs):
        """
        Performs an 'auth' request, which is to debit the money immediately
        as a one-off transaction.
        
        Note that currency should be ISO 4217 Alphabetic format.
        """ 
        self._check_kwargs(kwargs, ['amount', 'currency', 'card_number', 'expiry_date', 'merchant_reference'])
        response_xml = self._build_request_xml('auth', **kwargs)
        return self._build_response_dict(response_xml, {'authcode': 'auth_code'})
        
    def pre(self, **kwargs):
        """
        Performs an 'pre' request, which is to ring-fence the requested money
        so it can be fulfilled at a later time.
        """ 
        self._check_kwargs(kwargs, ['amount', 'currency', 'card_number', 'expiry_date', 'merchant_reference'])
        response_xml = self._build_request_xml('pre', **kwargs)
        return self._build_response_dict(response_xml, {'authcode': 'auth_code'})

    def refund(self, **kwargs):
        self._check_kwargs(kwargs, ['amount', 'currency', 'card_number', 'expiry_date', 'merchant_reference'])
        response_xml = self._build_request_xml('refund', **kwargs)
        return self._build_response_dict(response_xml, {'authcode': 'auth_code'})
        
    def erp(self, **kwargs):
        self._check_kwargs(kwargs, ['amount', 'currency', 'card_number', 'expiry_date', 'merchant_reference'])
        response_xml = self._build_request_xml('erp', **kwargs)
        return self._build_response_dict(response_xml, {'authcode': 'auth_code'})
        
    # "Historic" transaction types    
        
    def cancel(self, txn_reference): 
        response_xml = self._build_request_xml('cancel', txn_reference=txn_reference)
        return self._build_response_dict(response_xml)
    
    def fulfil(self, **kwargs):
        self._check_kwargs(kwargs, ['amount', 'currency', 'txn_reference', 'auth_code'])
        response_xml = self._build_request_xml('fulfil', **kwargs)
        return self._build_response_dict(response_xml)
    
    def txn_refund(self, **kwargs):
        self._check_kwargs(kwargs, ['amount', 'currency', 'txn_reference'])
        response_xml = self._build_request_xml('txn_refund', **kwargs)
        return self._build_response_dict(response_xml)
    
    def last_request_xml(self):
        return self._last_request_xml
    
    def last_response_xml(self):
        return self._last_response_xml
    
    def _check_kwargs(self, kwargs, required_keys):
        for key in required_keys:
            if key not in kwargs:
                raise RuntimeError('You must provide a "%s" argument' % key)


class Facade(object):
    """
    Responsible for dealing with oscar objects
    """
    
    def __init__(self):
        self.gateway = Gateway(settings.DATACASH_CLIENT, settings.DATACASH_PASSWORD, settings.DATACASH_HOST)
    
    def debit(self, order_number, amount, bankcard, basket, billing_address=None):
        with transaction.commit_on_success():
            response = self.gateway.auth(card_number=bankcard.card_number,
                                         expiry_date=bankcard.expiry_date,
                                         amount=amount,
                                         currency='GBP',
                                         merchant_reference=self.generate_merchant_reference(order_number),
                                         ccv=bankcard.ccv)
            
            # Create transaction model irrespective of whether transaction was successful or not
            txn = OrderTransaction.objects.create(order_number=order_number,
                                                  basket=basket,
                                                  method='auth',
                                                  datacash_ref=response['datacash_reference'],
                                                  merchant_ref=response['merchant_reference'],
                                                  amount=amount,
                                                  auth_code=response['auth_code'],
                                                  status=int(response['status']),
                                                  reason=response['reason'],
                                                  request_xml=self.gateway.last_request_xml(),
                                                  response_xml=self.gateway.last_response_xml())
        
        # Test if response is successful
        if response['status'] == INVALID_CREDENTIALS:
            # This needs to notify the administrators straight away
            import pprint
            msg = "Order #%s:\n%s" % (order_number, pprint.pprint(response))
            mail_admins("Datacash credentials are not valid", msg)
            raise InvalidGatewayRequestError("Unable to communicate with payment gateway, please try again later")
        
        if response['status'] == DECLINED:
            raise TransactionDeclined("Your bank declined this transaction, please check your details and try again")
        
        return response['datacash_reference']
        
    def generate_merchant_reference(self, order_number):
        return '%s_%s' % (order_number, datetime.datetime.now().microsecond)
        
        




import httplib

from django.test import TestCase
from django.core.exceptions import ValidationError
from django.core.urlresolvers import resolve
from django.test import Client

from oscar.apps.promotions.models import * 


class PromotionTest(TestCase):

    def test_default_template_name(self):
        promotion = Image.objects.create(name="dummy banner")
        self.assertEqual('promotions/image.html', promotion.template_name())





#class PagePromotionTest(unittest.TestCase):
#    
#    def setUp(self):
#        self.promotion = Promotion.objects.create(name='Dummy', link_url='http://www.example.com')
#        self.page_prom = PagePromotion.objects.create(promotion=self.promotion,
#                                                      position=RAW_HTML,
#                                                      page_url='/')
#    
#    def test_clicks_start_at_zero(self):
#        self.assertEquals(0, self.page_prom.clicks)
#    
#    def test_click_is_recorded(self):
#        self.page_prom.record_click()
#        self.assertEquals(1, self.page_prom.clicks)
#
#    def test_get_link(self):
#        link = self.page_prom.get_link()
#        match = resolve(link)
#        self.assertEquals('page-click', match.url_name)
#
#
#class KeywordPromotionTest(unittest.TestCase):
#    
#    def setUp(self):
#        self.promotion = Promotion.objects.create(name='Dummy', link_url='http://www.example.com')
#        self.kw_prom = KeywordPromotion.objects.create(promotion=self.promotion,
#                                                       position=RAW_HTML,
#                                                       keyword='cheese')
#    
#    def test_clicks_start_at_zero(self):
#        self.assertEquals(0, self.kw_prom.clicks)
#    
#    def test_click_is_recorded(self):
#        self.kw_prom.record_click()
#        self.assertEquals(1, self.kw_prom.clicks)
#        
#    def test_get_link(self):
#        link = self.kw_prom.get_link()
#        match = resolve(link)
#        self.assertEquals('keyword-click', match.url_name)
        



import csv

from oscar.core.loading import import_module
import_module('reports.reports', ['ReportGenerator'], locals())
import_module('order.models', ['Order'], locals())


class OrderReportGenerator(ReportGenerator):
    
    filename_template = 'orders-%s-to-%s.csv'
    code = 'order_report'
    description = "Orders placed"
    
    def generate(self, response):
        orders = Order._default_manager.filter(
            date_placed__gte=self.start_date
        ).filter(date_placed__lt=self.end_date)
        
        writer = csv.writer(response)
        header_row = ['Order number',
                      'User',
                      'Total incl. tax',
                      'Date placed',]
        writer.writerow(header_row)
        for order in orders:
            row = [order.number, order.user, order.total_incl_tax, order.date_placed]
            writer.writerow(row)
            
    def is_available_to(self, user):
        return user.is_staff and user.has_perm('order.can_view')


import os


class BaseCache(object):
    def __init__(self, path, config):
        self._path = path
        self._config = config

    def check(self, path):
        return False

    def write(self, data):
        pass

    def read(self):
        pass


class NullCache(BaseCache):
    """
    A dummy cache that doesn't do anything, handy for developing new Mods
    """
    def check(self, path):
        return False

    def write(self, data):
        self.data = data

    def read(self):
        return self.data


class DiskCache(BaseCache):
    """
    A simple disk-based cache.
    """
    def _create_folders(self):
        """
        Create the disk cache path so that the cached image can be stored in
        the same hierarchy as the original images.
        """
        paths = self._path.split(os.path.sep)
        paths.pop()  # Remove file from path
        path = os.path.join(self._config['cache_root'], *paths)

        if not os.path.isdir(path):
            os.makedirs(path)

    def _cache_path(self):
        return os.path.join(self._config['cache_root'], self._path)

    def check(self, path):
        """
        Checks the disk cache for an already processed image. If it exists then
        we'll check it's timestamp against the original image to make sure it's
        newer (and therefore valid). Also creates the folder hierarchy in the
        cache for the cached image if it doesn't find it there itself.
        """
        self._create_folders()

        cache = self._cache_path()

        original_time = os.path.getmtime(path)

        if os.path.exists(cache):
            cache_time = os.path.getmtime(cache)
        else:
            # Cached image does not exist
            return False

        if original_time > cache_time:
            # Cached image is out of date
            return False
        else:
            return True

    def write(self, data):
        path = self._cache_path()
        f = open(path, 'w')
        f.write(data)
        f.close()
        
    def file_info(self):
        """
        If we're using X-Sendfile or X-Accel-Redirect we want to return info
        about the file, rather than the actual file content
        """
        size = os.path.getsize(self._cache_path())
        return (self._path, size)

    def read(self):
        f = open(self._cache_path(), "r")
        data = f.read()
        f.close()

        return data



from django.conf.urls.defaults import patterns, url
from django.contrib.auth.decorators import login_required

from oscar.core.application import Application
from oscar.apps.checkout.views import IndexView, ShippingAddressView, UserAddressDeleteView, UserAddressCreateView, \
                                      UserAddressUpdateView, ShippingMethodView, PaymentMethodView, OrderPreviewView, \
                                      PaymentDetailsView, ThankYouView


class CheckoutApplication(Application):
    name = 'checkout'
    
    index_view = IndexView
    shipping_address_view = ShippingAddressView
    user_address_create_view = UserAddressCreateView
    user_address_update_view = UserAddressUpdateView
    user_address_delete_view = UserAddressDeleteView
    shipping_method_view = ShippingMethodView
    payment_method_view = PaymentMethodView
    order_preview_view = OrderPreviewView
    payment_details_view = PaymentDetailsView
    thankyou_view = ThankYouView

    def get_urls(self):
        urlpatterns = patterns('',
            url(r'^$', self.index_view.as_view(), name='index'),
            # Shipping/user address views
            url(r'shipping-address/$', self.shipping_address_view.as_view(), name='shipping-address'),
            url(r'user-address/create/$', login_required(self.user_address_create_view.as_view()), name='user-address-create'),
            url(r'user-address/edit/(?P<pk>\d+)/$', login_required(self.user_address_update_view.as_view()), name='user-address-update'),
            url(r'user-address/delete/(?P<pk>\d+)/$', login_required(self.user_address_delete_view.as_view()), name='user-address-delete'),
            # Shipping method views
            url(r'shipping-method/$', self.shipping_method_view.as_view(), name='shipping-method'),
            # Payment method views
            url(r'payment-method/$', self.payment_method_view.as_view(), name='payment-method'),
            url(r'preview/$', self.order_preview_view.as_view(), name='preview'),
            url(r'payment-details/$', self.payment_details_view.as_view(), name='payment-details'),
            url(r'thank-you/$', self.thankyou_view.as_view(), name='thank-you'),
        )
        return urlpatterns


application = CheckoutApplication()


from oscar.apps.customer.abstract_models import AbstractEmail, AbstractCommunicationEventType


class Email(AbstractEmail):
    pass


class CommunicationEventType(AbstractCommunicationEventType):
    pass


# -*- coding: utf-8 -*-

from django.contrib.auth.models import User
from django.test import TestCase

from oscar.apps.address.models import Country, UserAddress
from oscar.apps.order.models import ShippingAddress


class UserAddressTest(TestCase):
    
    def setUp(self):
        self.user = User.objects.create(username='dummy_user')
        self.country = Country.objects.get(iso_3166_1_a2='GB')
    
    def tearDown(self):
        self.user.delete()
    
    def test_titleless_salutation_is_stripped(self):
        a = UserAddress.objects.create(last_name='Barrington', line1="75 Smith Road", postcode="N4 8TY", 
                                       country=self.country, user=self.user)
        self.assertEquals("Barrington", a.salutation())
    
    def test_postcodes_are_saved_in_uppercase(self):
        a = UserAddress.objects.create(last_name='Barrington', line1="75 Smith Road", postcode="n4 8ty", 
                                       country=self.country, user=self.user)
        self.assertEquals("N4 8TY", a.postcode)   
        
    def test_fields_are_stripped_when_saved(self):
        a = UserAddress.objects.create(last_name='Barrington', line1="  75 Smith Road  ", postcode="  n4 8ty", 
                                       country=self.country, user=self.user)
        self.assertEquals("N4 8TY", a.postcode)      
        
    def test_active_address_fields_skips_whitespace_only_fields(self):
        a = UserAddress(first_name="   ", last_name='Barrington', line1="  75 Smith Road  ", postcode="  n4 8ty", 
                        country=self.country, user=self.user)
        active_fields = a.active_address_fields()
        self.assertEquals("Barrington", active_fields[0])   
        
    def test_summary_is_property(self):
        a = UserAddress(first_name=" Terry  ", last_name='Barrington', line1="  75 Smith Road  ", postcode="  n4 8ty", 
                        country=self.country, user=self.user)
        self.assertEquals("Terry Barrington, 75 Smith Road, N4 8TY, UNITED KINGDOM", a.summary)
        
    def test_populate_shipping_address_doesnt_set_id(self):
        a = UserAddress(first_name=" Terry  ", last_name='Barrington', line1="  75 Smith Road  ", postcode="  n4 8ty", 
                        country=self.country, user=self.user)
        sa = ShippingAddress()
        a.populate_alternative_model(sa)
        self.assertIsNone(sa.id)
        
    def test_populated_shipping_address_has_same_summary_user_address(self):
        a = UserAddress(first_name=" Terry  ", last_name='Barrington', line1="  75 Smith Road  ", postcode="  n4 8ty", 
                        country=self.country, user=self.user)
        sa = ShippingAddress()
        a.populate_alternative_model(sa)
        self.assertEquals(sa.summary, a.summary)
        
    def test_addresses_with_same_fields_ignoring_whitespace_have_same_hash(self):
        a1 = UserAddress(first_name=" Terry  ", last_name='Barrington', line1="  75 Smith Road  ", postcode="  n4 8ty", 
                         country=self.country, user=self.user)
        a2 = UserAddress(first_name=" Terry", last_name='   Barrington', line1="  75 Smith Road  ", postcode="N4 8ty", 
                         country=self.country, user=self.user)
        self.assertEquals(a1.generate_hash(), a2.generate_hash())
        
    def test_hashing_with_utf8(self):
        a = UserAddress(first_name=u"\u0141ukasz Smith", last_name=u'Smith', line1=u"75 Smith Road", postcode=u"n4 8ty", 
                        country=self.country, user=self.user)
        hash = a.active_address_fields()
        
    def test_city_is_alias_of_line4(self):
        a = UserAddress.objects.create(last_name='Barrington', line1="75 Smith Road", line4="London", postcode="n4 8ty", 
                                       country=self.country, user=self.user)
        self.assertEqual('London', a.city)
           



import string
import random

from django.contrib.auth.forms import AuthenticationForm
from django.utils.translation import ugettext_lazy as _
from django import forms
from django.contrib.auth.models import User

def generate_username():
    uname = ''.join([random.choice(string.letters + string.digits + '_') for i in range(30)]) 
    
    try:
        User.objects.get(username=uname)
        return generate_username()
    except User.DoesNotExist:
        return uname


class EmailAuthenticationForm(AuthenticationForm):
    """
    Extends the standard django AuthenticationForm, to support 75 character
    usernames. 75 character usernames are needed to support the EmailOrUsername
    auth backend.
    """
    username = forms.EmailField(label=_('Email Address'))


class EmailUserCreationForm(forms.ModelForm):
    email = forms.EmailField(label=_('Email Address'))
    password1 = forms.CharField(label=_('Password'), widget=forms.PasswordInput)
    password2 = forms.CharField(label=_('Confirm Password'), widget=forms.PasswordInput)
    
    class Meta:
        model = User
        fields = ('email', )
        
    def clean_email(self):
        email = self.cleaned_data['email']
        try:
            User.objects.get(email=email)
        except User.DoesNotExist:
            return email
        raise forms.ValidationError(_("A user with that email address already exists."))
    
    def clean_password2(self):
        password1 = self.cleaned_data.get('password1', '')
        password2 = self.cleaned_data.get('password2', '')
        
        if password1 != password2:
            raise forms.ValidationError(_("The two password fields didn't match."))
        return password2
    
    def save(self, commit=True):
        user = super(EmailUserCreationForm, self).save(commit=False)
        user.set_password(self.cleaned_data['password1'])
        user.username = generate_username()
        
        if commit:
            user.save()
        return user



from django.core import validators
from django.core.exceptions import ValidationError
from django.core.urlresolvers import resolve
from django.http import Http404
from django.utils.translation import ugettext_lazy as _


class ExtendedURLValidator(validators.URLValidator):
    def __call__(self, value):
        try:
            super(ExtendedURLValidator, self).__call__(value)
        except ValidationError:
            if value:
                self.validate_local_url(value)
            else:
                raise

    def validate_local_url(self, value):
        """
        Validate local URL name
        """
        try:
            value = self.fix_local_url(value)
            if self.verify_exists:
                resolve(value)
            self.is_local_url = True
        except Http404:
            raise ValidationError(_('Specified page does not exist'))

    def fix_local_url(self, value):
        """
        Puts preceding and trailing slashes to local URL name
        """
        if value != '/':
            value = '/' + value.strip('/') + '/'
        return value



from haystack.indexes import *

from oscar.core.loading import import_module
product_models = import_module('catalogue.models', ['Product'])


class AbstractProductIndex(SearchIndex):
    u"""
    Base class for products solr index definition.  Overide by creating your
    own copy of oscar.search_indexes.py
    """
    text = EdgeNgramField(document=True, use_template=True, template_name='search/indexes/product/item_text.txt')
    title = EdgeNgramField(model_attr='title', null=True)
    upc = CharField(model_attr="upc", null=True)
    product_score = FloatField(model_attr="score")
    date_created = DateTimeField(model_attr='date_created')
    date_updated = DateTimeField(model_attr='date_updated')

    def index_queryset(self):
        """
        Used when the entire index for model is updated.

        Orders by the most recently updated so that new objects are indexed first
        """
        return product_models.Product.objects.order_by('-date_updated')

    def get_updated_field(self):
        u"""
        Used to specify the field used to determine if an object has been updated

        Can be used to filter the query set when updating the index
        """
        return 'date_updated'





from decimal import Decimal
import datetime

from django.core import exceptions
from django.db import models
from django.utils.translation import ugettext as _


class AbstractVoucher(models.Model):
    """
    A voucher.  This is simply a link to a collection of offers

    Note that there are three possible "usage" models:
    (a) Single use
    (b) Multi-use
    (c) Once per customer
    """
    name = models.CharField(_("Name"), max_length=128,
        help_text=_("""This will be shown in the checkout and basket once the voucher is entered"""))
    code = models.CharField(_("Code"), max_length=128, db_index=True, unique=True,
        help_text=_("""Case insensitive / No spaces allowed"""))
    offers = models.ManyToManyField('offer.ConditionalOffer', related_name='vouchers', 
                                    limit_choices_to={'offer_type': "Voucher"})

    SINGLE_USE, MULTI_USE, ONCE_PER_CUSTOMER = ('Single use', 'Multi-use', 'Once per customer')
    USAGE_CHOICES = (
        (SINGLE_USE, _("Can be used once by one customer")),
        (MULTI_USE, _("Can be used multiple times by multiple customers")),
        (ONCE_PER_CUSTOMER, _("Can only be used once per customer")),
    )
    usage = models.CharField(_("Usage"), max_length=128, choices=USAGE_CHOICES, default=MULTI_USE)

    start_date = models.DateField()
    end_date = models.DateField()

    # Summary information
    num_basket_additions = models.PositiveIntegerField(default=0)
    num_orders = models.PositiveIntegerField(default=0)
    total_discount = models.DecimalField(decimal_places=2, max_digits=12, default=Decimal('0.00'))
    
    date_created = models.DateField(auto_now_add=True)

    class Meta:
        get_latest_by = 'date_created'
        abstract = True

    def __unicode__(self):
        return self.name

    def clean(self):
        if self.start_date and self.end_date and self.start_date > self.end_date:
            raise exceptions.ValidationError(_('End date should be later than start date'))

    def save(self, *args, **kwargs):
        self.code = self.code.upper()
        super(AbstractVoucher, self).save(*args, **kwargs)

    def is_active(self, test_date=None):
        """
        Tests whether this voucher is currently active.
        """
        if not test_date:
            test_date = datetime.date.today()
        return self.start_date <= test_date and test_date < self.end_date

    def is_available_to_user(self, user=None):
        """
        Tests whether this voucher is available to the passed user.
        
        Returns a tuple of a boolean for whether it is successulf, and a message
        """
        is_available, message = False, ''
        if self.usage == self.SINGLE_USE:
            is_available = self.applications.count() == 0
            if not is_available:
                message = "This voucher has already been used"
        elif self.usage == self.MULTI_USE:
            is_available = True
        elif self.usage == self.ONCE_PER_CUSTOMER:
            if not user.is_authenticated():
                is_available = False
                message = "This voucher is only available to signed in users"
            else:
                is_available = self.applications.filter(voucher=self, user=user).count() == 0
                if not is_available:
                    message = "You have already used this voucher in a previous order"
        return is_available, message
    
    def record_usage(self, order, user):
        """
        Records a usage of this voucher in an order.
        """
        self.applications.create(voucher=self, order=order, user=user)
        
        
class AbstractVoucherApplication(models.Model):
    """
    For tracking how often a voucher has been used
    """
    voucher = models.ForeignKey('voucher.Voucher', related_name="applications")
    # It is possible for an anonymous user to apply a voucher so we need to allow
    # the user to be nullable
    user = models.ForeignKey('auth.User', blank=True, null=True)
    order = models.ForeignKey('order.Order')
    date_created = models.DateField(auto_now_add=True)

    class Meta:
        abstract = True

    def __unicode__(self):
        return u"'%s' used by '%s'" % (self.voucher, self.user)        



import zlib

from django.conf import settings
from django.core.exceptions import SuspiciousOperation
from django.db.models import get_model

from oscar.core.loading import import_module
import_module('offer.utils', ['Applicator'], locals())
basket_model = get_model('basket', 'basket')


class BasketMiddleware(object):
    
    def process_request(self, request):
        self.cookies_to_delete = []
        basket = self.get_basket(request)
        self.apply_offers_to_basket(request, basket)   
        request.basket = basket
    
    def get_basket(self, request):  
        manager = basket_model.open
        cookie_basket = self.get_cookie_basket(settings.OSCAR_BASKET_COOKIE_OPEN, 
                                               request, manager)
        
        if request.user.is_authenticated():
            # Signed-in user: if they have a cookie basket too, it means
            # that they have just signed in and we need to merge their cookie
            # basket into their user basket, then delete the cookie
            try:
                basket, _ = manager.get_or_create(owner=request.user)
            except basket_model.MultipleObjectsReturned:
                # Not sure quite how we end up here with multiple baskets
                # We merge any  them and create a fresh one
                old_baskets = list(manager.filter(owner=request.user))
                basket = old_baskets[0]
                for other_basket in old_baskets[1:]:
                    self.merge_baskets(basket, other_basket)
                
            if cookie_basket:
                self.merge_baskets(basket, cookie_basket)
                self.cookies_to_delete.append(settings.OSCAR_BASKET_COOKIE_OPEN)
        elif cookie_basket:
            # Anonymous user with a basket tied to the cookie
            basket = cookie_basket
        else:
            # Anonymous user with no basket - we don't save the basket until
            # we need to.
            basket = basket_model() 
        return basket 
            
    def merge_baskets(self, master, slave):
        """
        Merge one basket into another.
        
        This is its own method to allow it to be overridden
        """
        master.merge(slave)    
        
    def process_response(self, request, response):
        
        # Delete any surplus cookies
        if hasattr(self, 'cookies_to_delete'):
            for cookie_key in self.cookies_to_delete:
                response.delete_cookie(cookie_key)
            
        # If a basket has had products added to it, but the user is anonymous
        # then we need to assign it to a cookie
        if hasattr(request, 'basket') and request.basket.id > 0 \
            and not request.user.is_authenticated() \
            and settings.OSCAR_BASKET_COOKIE_OPEN not in request.COOKIES:
            cookie = "%s_%s" % (request.basket.id, self.get_basket_hash(request.basket.id))
            response.set_cookie(settings.OSCAR_BASKET_COOKIE_OPEN, 
                                cookie, 
                                max_age=settings.OSCAR_BASKET_COOKIE_LIFETIME, 
                                httponly=True)
        return response
    
    def process_template_response(self, request, response):
        if response.context_data is None:
            response.context_data = {}
        response.context_data['basket'] = request.basket
        return response
    
    def get_cookie_basket(self, cookie_key, request, manager):
        """
        Looks for a basket which is referenced by a cookie.
        
        If a cookie key is found with no matching basket, then we add
        it to the list to be deleted.
        """
        basket = None
        if cookie_key in request.COOKIES:
            parts = request.COOKIES[cookie_key].split("_")
            if len(parts) != 2:
                return basket
            basket_id, basket_hash = parts
            if basket_hash == self.get_basket_hash(basket_id):
                try:
                    basket = basket_model.objects.get(pk=basket_id, owner=None)
                except basket_model.DoesNotExist:
                    self.cookies_to_delete.append(cookie_key)
            else:
                self.cookies_to_delete.append(cookie_key)
        return basket    
    
    def apply_offers_to_basket(self, request, basket):
        if not basket.is_empty:
            Applicator().apply(request, basket)
    
    def get_basket_hash(self, basket_id):
        return str(zlib.crc32(str(basket_id)+settings.SECRET_KEY))



from django.conf.urls.defaults import *
from haystack.query import SearchQuerySet

from oscar.core.loading import import_module
import_module('search.views', ['Suggestions', 'MultiFacetedSearchView'], locals())
import_module('search.forms', ['MultiFacetedSearchForm'], locals())
import_module('search.search_indexes', ['ProductIndex'], locals())


sqs = SearchQuerySet()
for field_name, field in ProductIndex.fields.items():
    if field.faceted is True:
        # Ensure we facet the results set by the defined facetable fields
        sqs.facet(field_name)

urlpatterns = patterns('search.apps.views',
    url(r'^suggest/$', Suggestions.as_view(), name='oscar-search-suggest'),
    url(r'^$', MultiFacetedSearchView(form_class=MultiFacetedSearchForm, 
                                      searchqueryset=sqs), name='oscar-search'),
)


from django.db.models.fields import CharField, DecimalField
from django.utils.translation import ugettext_lazy as _

from oscar.core import validators
from oscar.forms import fields

class ExtendedURLField(CharField):
    description = _("URL")

    def __init__(self, verbose_name=None, name=None, verify_exists=True, **kwargs):
        kwargs['max_length'] = kwargs.get('max_length', 200)
        CharField.__init__(self, verbose_name, name, **kwargs)
        validator = validators.ExtendedURLValidator(verify_exists=verify_exists)
        self.validators.append(validator)

    def formfield(self, **kwargs):
        # As with CharField, this will cause URL validation to be performed twice
        defaults = {
            'form_class': fields.ExtendedURLField,
        }
        defaults.update(kwargs)
        return super(ExtendedURLField, self).formfield(**defaults)


class PositiveDecimalField(DecimalField):
    def formfield(self, **kwargs):
        return super(PositiveDecimalField, self).formfield(min_value=0)



import logging
import sys
from optparse import make_option

from django.core.management.base import BaseCommand, CommandError
from django.contrib.auth.models import User

from oscar.core.loading import import_module
import_module('order.models', ['Order'], locals())
import_module('customer.models', ['CommunicationEventType'], locals())
import_module('customer.utils', ['Dispatcher'], locals())


class Command(BaseCommand):
    
    args = '<communication_event_type> <order number>'
    help = 'For testing the content of order emails'
    
    def handle(self, *args, **options):
        logger = self._get_logger()
        if len(args) != 2:
            raise CommandError("Please select a event type and order number")
        
        try:
            order = Order.objects.get(number=args[1])
        except Order.DoesNotExist:
            raise CommandError("No order found with number %s" % args[1])
        
        try:
            event_type = CommunicationEventType.objects.get(code=args[0])
        except CommunicationEventType.DoesNotExist:
            raise CommandError("No event type found with code %s" % args[0])
        
        messages = event_type.get_messages({'order': order})
        dispatcher = Dispatcher(logger)
        user = User(username='dummy', email='user@example.com')
        email = dispatcher.send_email_messages(user, messages)
        
        print "Subject: %s\nBody:\n\n%s"% (email.subject, email.body)
        
    def _get_logger(self):
        logger = logging.getLogger(__name__)
        stream = logging.StreamHandler(self.stdout)
        logger.addHandler(stream)
        logger.setLevel(logging.DEBUG)
        return logger
        
        
 
        
        
        
            


from decimal import Decimal
import zlib
import datetime

from django.db import models
from django.utils.translation import ugettext as _
from django.core.exceptions import ObjectDoesNotExist, PermissionDenied

from oscar.apps.basket.managers import OpenBasketManager, SavedBasketManager

# Basket statuses
# - Frozen is for when a basket is in the process of being submitted
#   and we need to prevent any changes to it.
OPEN, MERGED, SAVED, FROZEN, SUBMITTED = ("Open", "Merged", "Saved", "Frozen", "Submitted")


class AbstractBasket(models.Model):
    """
    Basket object
    """
    # Baskets can be anonymously owned (which are merged if the user signs in)
    owner = models.ForeignKey('auth.User', related_name='baskets', null=True)
    STATUS_CHOICES = (
        (OPEN, _("Open - currently active")),
        (MERGED, _("Merged - superceded by another basket")),
        (SAVED, _("Saved - for items to be purchased later")),
        (FROZEN, _("Frozen - the basket cannot be modified")),
        (SUBMITTED, _("Submitted - has been ordered at the checkout")),
    )
    status = models.CharField(_("Status"), max_length=128, default=OPEN, choices=STATUS_CHOICES)
    vouchers = models.ManyToManyField('voucher.Voucher', null=True)

    date_created = models.DateTimeField(auto_now_add=True)
    date_merged = models.DateTimeField(null=True, blank=True)
    date_submitted = models.DateTimeField(null=True, blank=True)

    class Meta:
        abstract = True

    objects = models.Manager()
    open = OpenBasketManager()
    saved = SavedBasketManager()

    def __init__(self, *args, **kwargs):
        super(AbstractBasket, self).__init__(*args, **kwargs)
        self._lines = None  # Cached queryset of lines
        self.discounts = None  # Dictionary of discounts
        self.exempt_from_tax = False

    def __unicode__(self):
        return u"%s basket (owner: %s, lines: %d)" % (self.status, self.owner, self.num_lines)

    def all_lines(self):
        """
        Return a cached set of basket lines.

        This is important for offers as they alter the line models and you don't
        want to reload them from the DB.
        """
        if self._lines is None:
            self._lines = self.lines.all()
        return self._lines

    # ============
    # Manipulation
    # ============

    def flush(self):
        """Remove all lines from basket."""
        if self.status == FROZEN:
            raise PermissionDenied("A frozen basket cannot be flushed")
        self.lines_all().delete()
        self._lines = None

    def add_product(self, item, quantity=1, options=None):
        """
        Convenience method for adding products to a basket

        The 'options' list should contains dicts with keys 'option' and 'value'
        which link the relevant product.Option model and string value respectively.
        """
        if options is None:
            options = []
        if not self.id:
            self.save()
        line_ref = self._create_line_reference(item, options)
        line, created = self.lines.get_or_create(line_reference=line_ref,
                                                 product=item,
                                                 defaults={'quantity': quantity})
        if created:
            for option_dict in options:
                line.attributes.create(option=option_dict['option'],
                                       value=option_dict['value'])
        else:
            line.quantity += quantity
            line.save()
        self._lines = None

    def get_discounts(self):
        if self.discounts is None:
            self.discounts = []
        return self.discounts

    def set_discounts(self, discounts):
        """
        Sets the discounts that apply to this basket.

        This should be a list of dictionaries
        """
        self.discounts = discounts

    def remove_discounts(self):
        """
        Remove any discounts so they get recalculated
        """
        self.discounts = []
        self._lines = None

    def merge_line(self, line):
        """
        For transferring a line from another basket to this one.

        This is used with the "Saved" basket functionality.
        """
        try:
            existing_line = self.lines.get(line_reference=line.line_reference)
        except ObjectDoesNotExist:
            # Line does not already exist - reassign its basket
            line.basket = self
            line.save()
        else:
            # Line already exists - bump its quantity and delete the old
            existing_line.quantity += line.quantity
            existing_line.save()
            line.delete()

    def merge(self, basket):
        """
        Merges another basket with this one.
        """
        for line_to_merge in basket.all_lines():
            self.merge_line(line_to_merge)
        basket.status = MERGED
        basket.date_merged = datetime.datetime.now()
        basket.save()
        self._lines = None

    def freeze(self):
        """
        Freezes the basket so it cannot be modified.
        """
        self.status = FROZEN
        self.save()

    def thaw(self):
        """
        Unfreezes a basket so it can be modified again
        """
        self.status = OPEN
        self.save()

    def set_as_submitted(self):
        """Mark this basket as submitted."""
        self.status = SUBMITTED
        self.date_submitted = datetime.datetime.now()
        self.save()

    def set_as_tax_exempt(self):
        self.exempt_from_tax = True
        for line in self.all_lines():
            line.set_as_tax_exempt()

    # =======
    # Helpers
    # =======

    def _create_line_reference(self, item, options):
        """
        Returns a reference string for a line based on the item
        and its options.
        """
        if not options:
            return item.id
        return "%d_%s" % (item.id, zlib.crc32(str(options)))

    def _get_total(self, property):
        """
        For executing a named method on each line of the basket
        and returning the total.
        """
        total = Decimal('0.00')
        for line in self.all_lines():
            try:
                total += getattr(line, property)
            except ObjectDoesNotExist:
                # Handle situation where the product may have been deleted
                pass
        return total

    # ==========
    # Properties
    # ==========

    @property
    def is_empty(self):
        """Return bool based on basket having 0 lines"""
        return self.num_lines == 0

    @property
    def total_excl_tax(self):
        """Return total line price excluding tax"""
        return self._get_total('line_price_excl_tax_and_discounts')

    @property
    def total_tax(self):
        """Return total tax for a line"""
        return self._get_total('line_tax')

    @property
    def total_incl_tax(self):
        """
        Return total price inclusive of tax and discounts
        """
        return self._get_total('line_price_incl_tax_and_discounts')

    @property
    def total_incl_tax_excl_discounts(self):
        """
        Return total price inclusive of tax but exclusive discounts
        """
        return self._get_total('line_price_incl_tax')

    @property
    def total_discount(self):
        return self._get_total('discount_value')

    @property
    def offer_discounts(self):
        """
        Return discounts from non-voucher sources.
        """
        offer_discounts = []
        for discount in self.get_discounts():
            if not discount['voucher']:
                offer_discounts.append(discount)
        return offer_discounts

    @property
    def voucher_discounts(self):
        """
        Return discounts from vouchers
        """
        voucher_discounts = []
        for discount in self.get_discounts():
            if discount['voucher']:
                voucher_discounts.append(discount)
        return voucher_discounts

    @property
    def total_excl_tax_excl_discounts(self):
        """
        Return total price excluding tax and discounts
        """
        return self._get_total('line_price_excl_tax')

    @property
    def num_lines(self):
        """Return number of lines"""
        return self.all_lines().count()

    @property
    def num_items(self):
        """Return number of items"""
        return reduce(lambda num,line: num+line.quantity, self.all_lines(), 0)

    @property
    def num_items_without_discount(self):
        """Return number of items"""
        num = 0
        for line in self.all_lines():
            num += line.quantity_without_discount
        return num

    @property
    def time_before_submit(self):
        if not self.date_submitted:
            return None
        return self.date_submitted - self.date_created

    @property
    def time_since_creation(self, test_datetime=None):
        if not test_datetime:
            test_datetime = datetime.datetime.now()
        return test_datetime - self.date_created

    def contains_voucher(self, code):
        """
        Test whether the basket contains a voucher with a given code
        """
        try:
            self.vouchers.get(code=code)
        except ObjectDoesNotExist:
            return False
        else:
            return True

class AbstractLine(models.Model):
    """
    A line of a basket (product and a quantity)
    """

    basket = models.ForeignKey('basket.Basket', related_name='lines')
    # This is to determine which products belong to the same line
    # We can't just use product.id as you can have customised products
    # which should be treated as separate lines.  Set as a
    # SlugField as it is included in the path for certain views.
    line_reference = models.SlugField(max_length=128, db_index=True)

    product = models.ForeignKey('catalogue.Product', related_name='basket_lines')
    quantity = models.PositiveIntegerField(default=1)

    # Track date of first addition
    date_created = models.DateTimeField(auto_now_add=True)

    # Instance variables used to persist discount information
    _discount = Decimal('0.00')
    _affected_quantity = 0
    _charge_tax = True

    class Meta:
        abstract = True
        unique_together = ("basket", "line_reference")

    def __unicode__(self):
        return u"%s, Product '%s', quantity %d" % (self.basket, self.product, self.quantity)

    def save(self, *args, **kwargs):
        """Saves a line or deletes if it's quanity is 0"""
        if self.basket.status not in (OPEN, SAVED):
            raise PermissionDenied("You cannot modify a %s basket" % self.basket.status.lower())
        if self.quantity == 0:
            return self.delete(*args, **kwargs)
        super(AbstractLine, self).save(*args, **kwargs)

    def set_as_tax_exempt(self):
        self._charge_tax = False

    # =============
    # Offer methods
    # =============

    def clear_discount(self):
        """
        Remove any discounts from this line.
        """
        self._discount = Decimal('0.00')
        self._affected_quantity = 0

    def discount(self, discount_value, affected_quantity):
        self._discount += discount_value
        self._affected_quantity += int(affected_quantity)

    def consume(self, quantity):
        if quantity > self.quantity - self._affected_quantity:
            inc = self.quantity - self._affected_quantity
        else:
            inc = quantity
        self._affected_quantity += int(inc)

    def get_price_breakdown(self):
        """
        Returns a breakdown of line prices after discounts have
        been applied.
        """
        prices = []
        if not self.has_discount:
            prices.append((self.unit_price_incl_tax, self.unit_price_excl_tax, self.quantity))
        else:
            # Need to split the discount among the affected quantity
            # of products.
            item_incl_tax_discount = self._discount / int(self._affected_quantity)
            item_excl_tax_discount = item_incl_tax_discount * self._tax_ratio
            prices.append((self.unit_price_incl_tax - item_incl_tax_discount,
                           self.unit_price_excl_tax - item_excl_tax_discount,
                           self._affected_quantity))
            if self.quantity_without_discount:
                prices.append((self.unit_price_incl_tax, self.unit_price_excl_tax, self.quantity_without_discount))
        return prices

    # =======
    # Helpers
    # =======

    def _get_stockrecord_property(self, property):
        if not self.product.stockrecord:
            return Decimal('0.00')
        else:
            attr = getattr(self.product.stockrecord, property)
            if attr is None:
                attr = Decimal('0.00')
            return attr

    @property
    def _tax_ratio(self):
        if not self.unit_price_incl_tax:
            return 0
        return self.unit_price_excl_tax / self.unit_price_incl_tax

    # ==========
    # Properties
    # ==========

    @property
    def has_discount(self):
        return self.quantity > self.quantity_without_discount

    @property
    def quantity_without_discount(self):
        return int(self.quantity - self._affected_quantity)

    @property
    def is_available_for_discount(self):
        return self.quantity_without_discount > 0

    @property
    def discount_value(self):
        return self._discount

    @property
    def unit_price_excl_tax(self):
        """Return unit price excluding tax"""
        return self._get_stockrecord_property('price_excl_tax')

    @property
    def unit_price_incl_tax(self):
        """Return unit price including tax"""
        if not self._charge_tax:
            return self.unit_price_excl_tax
        return self._get_stockrecord_property('price_incl_tax')

    @property
    def unit_tax(self):
        """Return tax of a unit"""
        if not self._charge_tax:
            return Decimal('0.00')
        return self._get_stockrecord_property('price_tax')

    @property
    def line_price_excl_tax(self):
        """Return line price excluding tax"""
        return self.quantity * self.unit_price_excl_tax

    @property
    def line_price_excl_tax_and_discounts(self):
        return self.line_price_excl_tax - self._discount * self._tax_ratio

    @property
    def line_tax(self):
        """Return line tax"""
        return self.quantity * self.unit_tax

    @property
    def line_price_incl_tax(self):
        """Return line price including tax"""
        return self.quantity * self.unit_price_incl_tax

    @property
    def line_price_incl_tax_and_discounts(self):
        return self.line_price_incl_tax - self._discount

    @property
    def description(self):
        """Return product description"""
        d = str(self.product)
        ops = []
        for attribute in self.attributes.all():
            ops.append("%s = '%s'" % (attribute.option.name, attribute.value))
        if ops:
            d = "%s (%s)" % (d.decode('utf-8'), ", ".join(ops))
        return d


class AbstractLineAttribute(models.Model):
    """An attribute of a basket line"""
    line = models.ForeignKey('basket.Line', related_name='attributes')
    option = models.ForeignKey('catalogue.Option')
    value = models.CharField(_("Value"), max_length=255)

    class Meta:
        abstract = True





from decimal import Decimal as D

from django.conf import settings
from django.core.urlresolvers import reverse
from django.test import TestCase

from oscar.apps.basket.models import Basket, Line
from oscar.test.helpers import create_product, TwillTestCase


class ViewTest(TwillTestCase):

    def test_for_smoke(self):
        self.visit('basket:summary')
        self.assertResponseCodeIs(200)
        self.assertPageContains('Basket')
        self.assertPageTitleMatches('Oscar')


class BasketModelTest(TestCase):

    def setUp(self):
        self.basket = Basket.objects.create()
        self.dummy_product = create_product()

    def test_empty_baskets_have_zero_lines(self):
        self.assertTrue(Basket().num_lines == 0)

    def test_new_baskets_are_empty(self):
        self.assertTrue(Basket().is_empty)

    def test_basket_have_with_one_line(self):
        Line.objects.create(basket=self.basket, product=self.dummy_product)
        self.assertTrue(self.basket.num_lines == 1)

    def test_add_product_creates_line(self):
        self.basket.add_product(self.dummy_product)
        self.assertTrue(self.basket.num_lines == 1)

    def test_adding_multiproduct_line_returns_correct_number_of_items(self):
        self.basket.add_product(self.dummy_product, 10)
        self.assertEqual(self.basket.num_items, 10)


class BasketViewsTest(TestCase):

    def test_empty_basket_view(self):
        url = reverse('basket:summary')
        response = self.client.get(url)
        self.assertEquals(200, response.status_code)
        self.assertEquals(0, response.context['basket'].num_lines)

    def test_anonymous_add_to_basket_creates_cookie(self):
        dummy_product = create_product(price=D('10.00'))
        url = reverse('basket:add')
        post_params = {'product_id': dummy_product.id,
                       'action': 'add',
                       'quantity': 1}
        response = self.client.post(url, post_params)
        self.assertTrue('oscar_open_basket' in response.cookies)


class BasketThresholdTest(TestCase):

    def setUp(self):
        self._old_threshold = settings.OSCAR_MAX_BASKET_QUANTITY_THRESHOLD
        settings.OSCAR_MAX_BASKET_QUANTITY_THRESHOLD = 3

    def tearDown(self):
        settings.OSCAR_MAX_BASKET_QUANTITY_THRESHOLD = self._old_threshold

    def test_adding_more_than_threshold_raises(self):
        dummy_product = create_product(price=D('10.00'))
        url = reverse('basket:add')
        post_params = {'product_id': dummy_product.id,
                       'action': 'add',
                       'quantity': 2}
        response = self.client.post(url, post_params)
        self.assertTrue('oscar_open_basket' in response.cookies)
        post_params = {'product_id': dummy_product.id,
                       'action': 'add',
                       'quantity': 2}
        response = self.client.post(url, post_params)
        self.assertTrue('Your basket currently has 2 items.' in
                        response.cookies['messages'].value)



from django import template
from django.db.models import get_model

from oscar.apps.basket.forms import AddToBasketForm
register = template.Library()
product_model = get_model('catalogue', 'product')

@register.tag(name="basket_form")
def do_basket_form(parse, token):
    tokens = token.split_contents()
    
    if len(tokens) < 3 or tokens[2] != 'as':
        raise template.TemplateSyntaxError("%r tag uses the following syntax: {% basket_form product_var as form_var %}" % tokens[0])
    
    product_var = tokens[1]
    form_var = tokens[3]
    
    return BasketFormNode(product_var, form_var)


class BasketFormNode(template.Node):
    def __init__(self, product_var, form_var):
        self.product_var = template.Variable(product_var)
        self.form_var = form_var
        
    def render(self, context):
        try:
            product = self.product_var.resolve(context)
        except template.VariableDoesNotExist:
            return ''
        
        if isinstance(product, product_model):
            initial = {
                'product_id': product.id,
            }
            context[self.form_var] = AddToBasketForm(instance=product,initial=initial)
        return ''


from django.contrib import admin

from oscar.core.loading import import_module
import_module('order.models', ['Order', 'OrderNote', 'CommunicationEvent', 
                                        'BillingAddress', 'ShippingAddress', 'Line',
                                        'LinePrice', 'ShippingEvent', 'ShippingEventType', 
                                        'PaymentEvent', 'PaymentEventType', 'LineAttribute', 'OrderDiscount'], locals())

class OrderAdmin(admin.ModelAdmin):
    raw_id_fields = ['basket','user','billing_address','shipping_address', ]
    list_display = ('number', 'total_incl_tax', 'site', 'user', 'billing_address', 'date_placed')
    readonly_fields = ('number', 'total_incl_tax', 'total_excl_tax', 'shipping_incl_tax', 'shipping_excl_tax')

class LineAdmin(admin.ModelAdmin):
    list_display = ('order', 'product', 'quantity')

class LinePriceAdmin(admin.ModelAdmin):
    list_display = ('order', 'line', 'price_incl_tax', 'quantity')

class ShippingEventTypeAdmin(admin.ModelAdmin):
    list_display = ('name', 'is_required', 'sequence_number')
    exclude = ('code',)
    
class PaymentEventTypeAdmin(admin.ModelAdmin):
    exclude = ('code',)
    
class OrderNoteAdmin(admin.ModelAdmin):
    exclude = ('user',)
    
    def save_model(self, request, obj, form, change):
        if not change:
            obj.user = request.user
        obj.save()
        
class OrderDiscountAdmin(admin.ModelAdmin):
    readonly_fields = ('order' ,'offer', 'voucher', 'voucher_code', 'amount')
    list_display = ('order' ,'offer', 'voucher', 'voucher_code', 'amount')
    
admin.site.register(Order, OrderAdmin)
admin.site.register(ShippingAddress)
admin.site.register(Line, LineAdmin)
admin.site.register(LinePrice, LinePriceAdmin)
admin.site.register(ShippingEvent)
admin.site.register(ShippingEventType, ShippingEventTypeAdmin)
admin.site.register(PaymentEvent)
admin.site.register(PaymentEventType, PaymentEventTypeAdmin)
admin.site.register(LineAttribute)
admin.site.register(OrderDiscount, OrderDiscountAdmin)
admin.site.register(CommunicationEvent)





"""
Installation script:

To release a new version to PyPi:
- Ensure the version is correctly set in oscar.__init__.py
- Run: python setup.py sdist upload
"""

from setuptools import setup, find_packages

from oscar import get_version


setup(name='django-oscar',
      version=get_version().replace(' ', '-'),
      url='https://github.com/tangentlabs/django-oscar',
      author="Tangent Labs",
      author_email="david.winterbottom@tangentlabs.co.uk",
      description="A domain-driven e-commerce framework for Django 1.3+",
      long_description=open('README.rst').read(),
      keywords="E-commerce, Django, domain-driven",
      license='BSD',
      platforms=['linux'],
      packages=find_packages(exclude=["*.tests"]),
      install_requires=[
          'django-extra-views>=0.1.0',
          'django-haystack>=1.2.0',
          'django-treebeard>=1.6.1',
          'sorl-thumbnail>=11.05.1',
          ],
      # See http://pypi.python.org/pypi?%3Aaction=list_classifiers
      classifiers=['Environment :: Web Environment',
                   'Framework :: Django',
                   'Intended Audience :: Developers',
                   'License :: OSI Approved :: BSD License',
                   'Operating System :: Unix',
                   'Programming Language :: Python']
      )



from django.db import connection, transaction

from oscar.core.loading import import_module
import_module('analytics.models', ['ProductRecord'], locals())
import_module('catalogue.models', ['Product'], locals())


class ScoreCalculator(object):
    
    # Map of field name to weight
    weights = {'num_views': 1,
               'num_basket_additions': 3,
               'num_purchases': 5}
    
    def __init__(self, logger):
        self.logger = logger
        self.cursor = connection.cursor()
    
    def run(self):
        self.calculate_scores()
        self.update_product_models()
        
    def calculate_scores(self):
        self.logger.info("Calculating product scores")
        
        # Build the "SET ..." part of the SQL statement
        weighted_sum = " + ".join(["%s*`%s`" % (weight, field) for field, weight in self.weights.items()])
        
        ctx = {'table': ProductRecord._meta.db_table,
               'weighted_total': weighted_sum,
               'total_weight': sum(self.weights.values())}
        sql = '''UPDATE `%(table)s` 
                 SET score = (%(weighted_total)s) / %(total_weight)s''' % ctx
        self.logger.debug(sql)         
        self.cursor.execute(sql)
        transaction.commit_unless_managed()
        
    def update_product_models(self):
        self.logger.info("Updating product records")
        sql = '''UPDATE `%s` product, `%s` analytics
                 SET product.score = analytics.score
                 WHERE product.id = analytics.product_id''' % (Product._meta.db_table, ProductRecord._meta.db_table)
        self.cursor.execute(sql)
        transaction.commit_unless_managed()


from django import forms
from django.conf import settings
from django.db.models import get_model
from django.utils.translation import gettext_lazy as _

basketline_model = get_model('basket', 'line')
basket_model = get_model('basket', 'basket')
Product = get_model('catalogue', 'product')


class BasketLineForm(forms.ModelForm):
    save_for_later = forms.BooleanField(initial=False, required=False)

    class Meta:
        model = basketline_model
        exclude = ('basket', 'product', 'line_reference', )


class SavedLineForm(forms.ModelForm):
    move_to_basket = forms.BooleanField(initial=False, required=False)

    class Meta:
        model = basketline_model
        exclude = ('basket', 'product', 'line_reference', 'quantity', )


class BasketVoucherForm(forms.Form):
    code = forms.CharField(max_length=128)

    def __init__(self, *args, **kwargs):
        return super(BasketVoucherForm, self).__init__(*args,**kwargs)


class ProductSelectionForm(forms.Form):
    product_id = forms.IntegerField(min_value=1)

    def clean_product_id(self):
        id = self.cleaned_data['product_id']

        try:
            return Product.objects.get(pk=id)
        except Product.DoesNotExist:
            raise forms.ValidationError(_("This product is not available for purchase"))


class AddToBasketForm(forms.Form):
    product_id = forms.IntegerField(widget=forms.HiddenInput(), min_value=1)
    quantity = forms.IntegerField(initial=1, min_value=1)

    def __init__(self, basket, instance, *args, **kwargs):
        super(AddToBasketForm, self).__init__(*args, **kwargs)
        self.basket = basket
        self.instance = instance
        if instance:
            if instance.is_group:
                self._create_group_product_fields(instance)
            else:
                self._create_product_fields(instance)

    def clean_product_id(self):
        id = self.cleaned_data['product_id']
        product = Product.objects.get(id=id)
        if not product.has_stockrecord or not product.stockrecord.is_available_to_buy:
            raise forms.ValidationError(_("This product is not available for purchase"))
        return id

    def clean_quantity(self):
        qty = self.cleaned_data['quantity']
        basket_threshold = settings.OSCAR_MAX_BASKET_QUANTITY_THRESHOLD
        if basket_threshold:
            total_basket_quantity = self.basket.num_items
            max_allowed = basket_threshold - total_basket_quantity
            if qty > max_allowed:
                raise forms.ValidationError(
                    _("Due to technical limitations we are not able to ship"
                      " more than %(threshold)d items in one order. Your basket"
                      " currently has %(basket)d items.") % {
                            'threshold': basket_threshold,
                            'basket': total_basket_quantity,
                    })
        return qty

    def _create_group_product_fields(self, item):
        """
        Adds the fields for a "group"-type product (eg, a parent product with a
        list of variants.
        """
        choices = []
        for variant in item.variants.all():
            if variant.has_stockrecord:
                summary = u"%s (%s) - %.2f" % (variant.get_title(), variant.attribute_summary(),
                                               variant.stockrecord.price_incl_tax)
                choices.append((variant.id, summary))
        self.fields['product_id'] = forms.ChoiceField(choices=tuple(choices))

    def _create_product_fields(self, item):
        u"""Add the product option fields."""
        for option in item.options:
            self._add_option_field(item, option)

    def _add_option_field(self, item, option):
        u"""
        Creates the appropriate form field for the product option.

        This is designed to be overridden so that specific widgets can be used for
        certain types of options.
        """
        self.fields[option.code] = forms.CharField()





"""
Models of products
"""
import re
from itertools import chain
from datetime import datetime, date

from django.db import models
from django.core.validators import RegexValidator
from django.db.models import get_model
from django.conf import settings
from django.core.exceptions import ValidationError
from django.utils.translation import ugettext_lazy as _
from django.template.defaultfilters import slugify
from django.core.exceptions import ObjectDoesNotExist
from treebeard.mp_tree import MP_Node

from oscar.apps.catalogue.managers import BrowsableProductManager

ENABLE_ATTRIBUTE_BINDING = getattr(settings, 'OSCAR_ENABLE_ATTRIBUTE_BINDING', False)

class AbstractProductClass(models.Model):
    """
    Defines the options and attributes for a group of products, e.g. Books, DVDs and Toys.
    Not necessarily equivalent to top-level categories but usually will be.
    """
    name = models.CharField(_('name'), max_length=128)
    slug = models.SlugField(max_length=128, unique=True)
    
    # These are the options (set by the user when they add to basket) for this item class
    options = models.ManyToManyField('catalogue.Option', blank=True)

    class Meta:
        abstract = True
        ordering = ['name']
        verbose_name_plural = "Product classes"

    def save(self, *args, **kwargs):
        if not self.slug:
            self.slug= slugify(self.name)
        super(AbstractProductClass, self).save(*args, **kwargs)

    def __unicode__(self):
        return self.name


class AbstractCategory(MP_Node):
    """
    Category hierarchy, top-level nodes represent departments. Uses django-treebeard.
    """
    name = models.CharField(max_length=255, db_index=True)
    slug = models.SlugField(max_length=1024, db_index=True, editable=False)
    full_name = models.CharField(max_length=1024, db_index=True, editable=False)    
    
    def __unicode__(self):
        return self.name
    
    def save(self, *args, **kwargs):
        if not self.slug:
            parent = self.get_parent()
            slug = slugify(self.name)
            if parent:
                self.slug = '%s/%s' % (parent.slug, slug)
                self.full_name = '%s > %s' % (parent.full_name, self.name)
            else:
                self.slug = slug
                self.full_name = self.name
        super(AbstractCategory, self).save(*args, **kwargs)

    def get_ancestors(self, include_self=True):
        ancestors = list(super(AbstractCategory, self).get_ancestors())
        if include_self:
            ancestors.append(self)
        return ancestors

    @models.permalink
    def get_absolute_url(self):
        return ('catalogue:category', (), {
                'category_slug': self.slug })

    class Meta:
        abstract = True
        ordering = ['name']
        verbose_name_plural = 'Categories'
        verbose_name = 'Category'


class AbstractProductCategory(models.Model):
    """
    Joining model between products and categories.
    """
    product = models.ForeignKey('catalogue.Product')
    category = models.ForeignKey('catalogue.Category')
    is_canonical = models.BooleanField(default=False, db_index=True)
    
    class Meta:
        abstract = True
        ordering = ['-is_canonical']
        verbose_name_plural = 'Categories'
        

class AbstractContributorRole(models.Model):
    """
    A role that may be performed by a contributor to a product, eg Author, Actor, Director.
    """
    name = models.CharField(max_length=50)
    name_plural = models.CharField(max_length=50)
    slug = models.SlugField()
    
    def __unicode__(self):
        return self.name
    
    class Meta:
        abstract = True
        
    def save(self, *args, **kwargs):
        if not self.slug:
            self.slug = slugify(self.name)
        super(AbstractContributorRole, self).save(*args, **kwargs) 


class AbstractContributor(models.Model):
    """
    Represents a person or business that has contributed to a product in some way. eg an author.
    """
    name = models.CharField(_("Name"), max_length=255)
    slug = models.SlugField(max_length=255, unique=False)

    def __unicode__(self):
        return self.name
    
    class Meta:
        abstract = True
        
    def save(self, *args, **kwargs):
        if not self.slug:
            self.slug = slugify(self.name)
        super(AbstractContributor, self).save(*args, **kwargs)


class AbstractProductContributor(models.Model):
    product = models.ForeignKey('catalogue.Product')
    contributor = models.ForeignKey('catalogue.Contributor')
    role = models.ForeignKey('catalogue.ContributorRole')
    
    def __unicode__(self):
        return '%s <- %s - %s' % (self.product, self.role, self.contributor)
    
    class Meta:
        abstract = True


class AbstractProduct(models.Model):
    """
    The base product object
    """
    # If an item has no parent, then it is the "canonical" or abstract version of a product
    # which essentially represents a set of products.  If a product has a parent
    # then it is a specific version of a catalogue.
    # 
    # For example, a canonical product would have a title like "Green fleece" while its 
    # children would be "Green fleece - size L".
    
    # Universal product code
    upc = models.CharField(_("UPC"), max_length=64, blank=True, null=True, db_index=True,
        help_text="""Universal Product Code (UPC) is an identifier for a product which is 
                     not specific to a particular supplier.  Eg an ISBN for a book.""")
    
    # No canonical product should have a stock record as they cannot be bought.
    parent = models.ForeignKey('self', null=True, blank=True, related_name='variants',
        help_text="""Only choose a parent product if this is a 'variant' of a canonical catalogue.  For example 
                     if this is a size 4 of a particular t-shirt.  Leave blank if this is a CANONICAL PRODUCT (ie 
                     there is only one version of this product).""")
    
    # Title is mandatory for canonical products but optional for child products
    title = models.CharField(_('Title'), max_length=255, blank=True, null=True)
    slug = models.SlugField(max_length=255, unique=False)
    description = models.TextField(_('Description'), blank=True, null=True)
    product_class = models.ForeignKey('catalogue.ProductClass', verbose_name=_('product class'), null=True,
        help_text="""Choose what type of product this is""")
    attributes = models.ManyToManyField('catalogue.ProductAttribute', through='ProductAttributeValue',
        help_text="""A product attribute is something that this product MUST have, such as a size, as specified by its class""")
    product_options = models.ManyToManyField('catalogue.Option', blank=True, 
        help_text="""Options are values that can be associated with a item when it is added to 
                     a customer's basket.  This could be something like a personalised message to be
                     printed on a T-shirt.<br/>""")
    
    related_products = models.ManyToManyField('catalogue.Product', related_name='relations', blank=True, help_text="""Related 
        items are things like different formats of the same book.  Grouping them together allows
        better linking betwen products on the site.<br/>""")
    
    # Recommended products
    recommended_products = models.ManyToManyField('catalogue.Product', through='ProductRecommendation', blank=True)
    
    # Product score
    score = models.FloatField(default=0.00, db_index=True)
    
    date_created = models.DateTimeField(auto_now_add=True)

    # This field is used by Haystack to reindex search
    date_updated = models.DateTimeField(auto_now=True, db_index=True)
    
    categories = models.ManyToManyField('catalogue.Category', through='ProductCategory')

    objects = models.Manager()
    browsable = BrowsableProductManager()

    # Properties

    @property
    def options(self):
        return list(chain(self.product_options.all(), self.get_product_class().options.all()))

    @property
    def is_top_level(self):
        u"""Return True if this is a parent product"""
        return self.parent_id == None
    
    @property
    def is_group(self):
        u"""Return True if this is a top level product and has more than 0 variants"""
        return self.is_top_level and self.variants.count() > 0
    
    @property
    def is_variant(self):
        u"""Return True if a product is not a top level product"""
        return not self.is_top_level

    @property
    def min_variant_price_incl_tax(self):
        u"""Return minimum variant price including tax"""
        return self._min_variant_price('price_incl_tax')
    
    @property
    def min_variant_price_excl_tax(self):
        u"""Return minimum variant price excluding tax"""
        return self._min_variant_price('price_excl_tax')

    @property
    def has_stockrecord(self):
        u"""Return True if a product has a stock record, False if not"""
        try:
            self.stockrecord
            return True
        except ObjectDoesNotExist:
            return False

    def add_category_from_breadcrumbs(self, breadcrumb):
        from oscar.apps.catalogue.utils import breadcrumbs_to_category
        category = breadcrumbs_to_category(breadcrumb)
        
        temp = models.get_model('product', 'productcategory')(category=category, product=self)
        temp.save()

    def attribute_summary(self):
        u"""Return a string of all of a product's attributes"""
        return ", ".join([attribute.__unicode__() for attribute in self.attributes.all()])

    def get_title(self):
        u"""Return a product's title or it's parent's title if it has no title"""
        title = self.__dict__.setdefault('title', '')
        if not title and self.parent_id:
            title = self.parent.title
        return title
    
    def get_product_class(self):
        u"""Return a product's item class"""
        if self.product_class:
            return self.product_class
        if self.parent.product_class:
            return self.parent.product_class
        return None

    # Helpers
    
    def _min_variant_price(self, property):
        u"""Return minimum variant price"""
        prices = []
        for variant in self.variants.all():
            if variant.has_stockrecord:
                prices.append(getattr(variant.stockrecord, property))
        if not prices:
            return None
        prices.sort()
        return prices[0]

    class Meta:
        abstract = True
        ordering = ['-date_created']

    def __unicode__(self):
        if self.is_variant:
            return u"%s (%s)" % (self.get_title(), self.attribute_summary())
        return self.get_title()
    
    @models.permalink
    def get_absolute_url(self):
        u"""Return a product's absolute url"""
        return ('catalogue:detail', (), {
            'product_slug': self.slug,
            'pk': self.id})
        
    def __init__(self, *args, **kwargs):
        super(AbstractProduct, self).__init__(*args, **kwargs)
        if ENABLE_ATTRIBUTE_BINDING:
            self.attr = ProductAttributesContainer(product=self)
    
    def save(self, *args, **kwargs):
        if self.is_top_level and not self.title:
            raise ValidationError("Canonical products must have a title")
        if not self.slug:
            self.slug = slugify(self.get_title())
        
        # Validate attributes if necessary
        if ENABLE_ATTRIBUTE_BINDING:
            self.attr.validate_attributes()
            
        # Save product
        super(AbstractProduct, self).save(*args, **kwargs)
        
        # Finally, save attributes
        if ENABLE_ATTRIBUTE_BINDING:
            self.attr.save()


class ProductRecommendation(models.Model):
    u"""
    'Through' model for product recommendations
    """
    primary = models.ForeignKey('catalogue.Product', related_name='primary_recommendations')
    recommendation = models.ForeignKey('catalogue.Product')
    ranking = models.PositiveSmallIntegerField(default=0)


class ProductAttributesContainer(object):
    """
    Stolen liberally from django-eav, but simplified to be product-specific
    """
    
    def __init__(self, product):
        self.product = product
        self.initialised = False

    def __getattr__(self, name):
        if not name.startswith('_') and not self.initialised:
            values = list(self.get_values().select_related('attribute'))
            result = None
            for v in values:
                setattr(self, v.attribute.code, v.value)
                if v.attribute.code == name:
                    result = v.value
            self.initialised = True
            if result:
                return result
        raise AttributeError((_(u"%(obj)s has no attribute named " \
                                       u"'%(attr)s'") % \
                                     {'obj': self.product.product_class, 'attr': name}))
        
    def validate_attributes(self):
        for attribute in self.get_all_attributes():
            value = getattr(self, attribute.code, None)
            if value is None:
                if attribute.required:
                    raise ValidationError(_(u"%(attr)s attribute cannot " \
                                            u"be blank") % \
                                            {'attr': attribute.code})
            else:
                try:
                    attribute.validate_value(value)
                except ValidationError, e:
                    raise ValidationError(_(u"%(attr)s attribute %(err)s") % \
                                            {'attr': attribute.code,
                                             'err': e})
        
    def get_values(self):
        return self.product.productattributevalue_set.all()
    
    def get_value_by_attribute(self, attribute):
        return self.get_values().get(attribute=attribute)    
    
    def get_all_attributes(self):
        return self.product.product_class.attributes.all()
    
    def get_attribute_by_code(self, code):
        return self.get_all_attributes().get(code=code)
    
    def __iter__(self):
        return iter(self.get_values())
    
    def save(self):
        for attribute in self.get_all_attributes():
            if hasattr(self, attribute.code):
                value = getattr(self, attribute.code)
                attribute.save_value(self.product, value)


class AbstractProductAttribute(models.Model):
    
    TYPE_CHOICES = (
        ("text", "Text"), 
        ("integer", "Integer"),
        ("boolean", "True / False"), 
        ("float", "Float"), 
        ("richtext", "Rich Text"), 
        ("date", "Date"), 
        ("option", "Option"),
        ("entity", "Entity"),
    )
    
    """
    Defines an attribute for a product class. (For example, number_of_pages for a 'book' class)
    """
    product_class = models.ForeignKey('catalogue.ProductClass', related_name='attributes', blank=True, null=True)
    name = models.CharField(_('name'), max_length=128)
    code = models.SlugField(_('code'), max_length=128, validators=[RegexValidator(regex=r'^[a-zA-Z_][0-9a-zA-Z_]*$', message="Code must match ^[a-zA-Z_][0-9a-zA-Z_]*$")])
    type = models.CharField(choices=TYPE_CHOICES, default=TYPE_CHOICES[0][0], max_length=20)
    option_group = models.ForeignKey('catalogue.AttributeOptionGroup', blank=True, null=True, help_text='Select an option group if using type "Option"')
    entity_type = models.ForeignKey('catalogue.AttributeEntityType', blank=True, null=True, help_text='Select an entity type if using type "Entity"')
    required = models.BooleanField(default=False)

    class Meta:
        abstract = True 
        ordering = ['code']

    def _validate_text(self, value):
        if not (type(value) == unicode or type(value) == str):
            raise ValidationError(_(u"Must be str or unicode"))

    def _validate_float(self, value):
        try:
            float(value)
        except ValueError:
            raise ValidationError(_(u"Must be a float"))

    def _validate_int(self, value):
        try:
            int(value)
        except ValueError:
            raise ValidationError(_(u"Must be an integer"))

    def _validate_date(self, value):
        if not (isinstance(value, datetime) or isinstance(value, date)):
            raise ValidationError(_(u"Must be a date or datetime"))

    def _validate_bool(self, value):
        if not type(value) == bool:
            raise ValidationError(_(u"Must be a boolean"))

    def _validate_entity(self, value):
        if not isinstance(value, get_model('catalogue', 'AttributeEntity')):
            raise ValidationError(_(u"Must be an AttributeEntity model object instance"))
        if not value.pk:
            raise ValidationError(_(u"Model has not been saved yet"))
        if value.type != self.entity_type:
            raise ValidationError(_(u"Entity must be of type %s" % self.entity_type.name))

    def _validate_option(self, value):
        if not isinstance(value, get_model('catalogue', 'AttributeOption')):
            raise ValidationError(_(u"Must be an AttributeOption model object instance"))
        if not value.pk:
            raise ValidationError(_(u"AttributeOption has not been saved yet"))
        valid_values = self.option_group.options.values_list('option', flat=True)
        if value not in valid_values:
            raise ValidationError(_(u"%(enum)s is not a valid choice "
                                        u"for %(attr)s") % \
                                       {'enum': value, 'attr': self})        


    def get_validator(self):
        DATATYPE_VALIDATORS = {
            'text': self._validate_text,
            'integer': self._validate_int,
            'boolean': self._validate_bool,
            'float': self._validate_float,
            'richtext': self._validate_text,
            'date': self._validate_date,
            'entity': self._validate_entity,
            'option': self._validate_option,
        }

        return DATATYPE_VALIDATORS[self.type]     

    def __unicode__(self):
        return self.name

    def save(self, *args, **kwargs):
        super(AbstractProductAttribute, self).save(*args, **kwargs)
        
    def save_value(self, product, value):
        try:
            value_obj = product.productattributevalue_set.get(attribute=self)
        except get_model('catalogue', 'ProductAttributeValue').DoesNotExist:
            if value == None or value == '':
                return
            value_obj = get_model('catalogue', 'ProductAttributeValue').objects.create(product=product, attribute=self)
        if value == None or value == '':
            value_obj.delete()
            return
        if value != value_obj.value:
            value_obj.value = value
            value_obj.save()
    
    def validate_value(self, value):
        self.get_validator()(value)
        
    def is_value_valid(self, value):
        """
        Check whether the passed value is valid for this attribute
        """
        if self.type == 'option':
            valid_values = self.option_group.options.values_list('option', flat=True)
            return value in valid_values
        return True


class AbstractProductAttributeValue(models.Model):
    """
    The "through" model for the m2m relationship between catalogue.Product
    and catalogue.ProductAttribute.  
    This specifies the value of the attribute for a particular product
    
    For example: number_of_pages = 295
    """
    attribute = models.ForeignKey('catalogue.ProductAttribute')
    product = models.ForeignKey('catalogue.Product')
    value_text = models.CharField(max_length=255, blank=True, null=True)
    value_integer = models.IntegerField(blank=True, null=True)
    value_boolean = models.BooleanField(blank=True)
    value_float = models.FloatField(blank=True, null=True)
    value_richtext = models.TextField(blank=True, null=True)
    value_date = models.DateField(blank=True, null=True)
    value_option = models.ForeignKey('catalogue.AttributeOption', blank=True, null=True)
    value_entity = models.ForeignKey('catalogue.AttributeEntity', blank=True, null=True)
    
    def _get_value(self):
        return getattr(self, 'value_%s' % self.attribute.type)
    
    def _set_value(self, new_value):
        if self.attribute.type == 'option' and isinstance(new_value, str):
            # Need to look up instance of AttributeOption
            new_value = self.attribute.option_group.options.get(option=new_value)
        setattr(self, 'value_%s' % self.attribute.type, new_value)
    
    value = property(_get_value, _set_value)
    
    class Meta:
        abstract = True 
        
    def __unicode__(self):
        return u"%s: %s" % (self.attribute.name, self.value)
    
    
class AbstractAttributeOptionGroup(models.Model):
    """
    Defines a group of options that collectively may be used as an 
    attribute type
    For example, Language
    """
    name = models.CharField(max_length=128)
    
    def __unicode__(self):
        return self.name
    
    class Meta:
        abstract = True


class AbstractAttributeOption(models.Model):
    """
    Provides an option within an option group for an attribute type
    Examples: In a Language group, English, Greek, French
    """
    group = models.ForeignKey('catalogue.AttributeOptionGroup', related_name='options')
    option = models.CharField(max_length=255)
    
    def __unicode__(self):
        return self.option
    
    class Meta:
        abstract = True
    
    
class AbstractAttributeEntity(models.Model):
    """
    Provides an attribute type to enable relationships with other models
    """
    name = models.CharField(_("Name"), max_length=255)
    slug = models.SlugField(max_length=255, unique=False, blank=True)
    type = models.ForeignKey('catalogue.AttributeEntityType', related_name='entities')

    def __unicode__(self):
        return self.name
    
    class Meta:
        abstract = True
        verbose_name_plural = 'Attribute entities'
        
    def save(self, *args, **kwargs):
        if not self.slug:
            self.slug = slugify(self.name)
        super(AbstractAttributeEntity, self).save(*args, **kwargs)


class AbstractAttributeEntityType(models.Model):
    """
    Provides the name of the model involved in an entity relationship
    """
    name = models.CharField(_("Name"), max_length=255)
    slug = models.SlugField(max_length=255, unique=False, blank=True)
    
    def __unicode__(self):
        return self.name
        
    class Meta:
        abstract = True
        
    def save(self, *args, **kwargs):
        if not self.slug:
            self.slug = slugify(self.name)
        super(AbstractAttributeEntityType, self).save(*args, **kwargs)
    
    
class AbstractOption(models.Model):
    u"""
    An option that can be selected for a particular item when the product
    is added to the basket.  
    
    Eg a list ID for an SMS message send, or a personalised message to 
    print on a T-shirt.  
    
    This is not the same as an attribute as options do not have a fixed value for 
    a particular item - options, they need to be specified by the customer.
    """
    name = models.CharField(_('name'), max_length=128)
    code = models.SlugField(_('code'), max_length=128)
    
    REQUIRED, OPTIONAL = ('Required', 'Optional')
    TYPE_CHOICES = (
        (REQUIRED, _("Required - a value for this option must be specified")),
        (OPTIONAL, _("Optional - a value for this option can be omitted")),
    )
    type = models.CharField(_("Status"), max_length=128, default=REQUIRED, choices=TYPE_CHOICES)
    
    class Meta:
        abstract = True
        
    def __unicode__(self):
        return self.name
    
    def save(self, *args, **kwargs):
        if not self.code:
            self.code = slugify(self.name)
        super(AbstractOption, self).save(*args, **kwargs)


class AbstractProductImage(models.Model):
    u"""An image of a product"""
    product = models.ForeignKey('catalogue.Product', related_name='images')
    
    original = models.ImageField(upload_to=settings.OSCAR_IMAGE_FOLDER)
    caption = models.CharField(_("Caption"), max_length=200, blank=True, null=True)
    
    # Use display_order to determine which is the "primary" image
    display_order = models.PositiveIntegerField(default=0, help_text="""An image with a display order of
       zero will be the primary image for a product""")
    date_created = models.DateTimeField(auto_now_add=True)
    
    class Meta:
        abstract = True
        unique_together = ("product", "display_order")
        ordering = ["display_order"]

    def __unicode__(self):
        return u"Image of '%s'" % self.product

    def is_primary(self):
        u"""Return bool if image display order is 0"""
        return self.display_order == 0

    def resized_image_url(self, width=None, height=None, **kwargs):
        return self.original.url

    def fullsize_url(self):
        u"""
        Returns the URL path for this image.  This is intended
        to be overridden in subclasses that want to serve
        images in a specific way.
        """
        return self.resized_image_url()
    
    def thumbnail_url(self):
        return self.resized_image_url()




from oscar.core.loading import import_module
order_reports = import_module('order.reports', ['OrderReportGenerator'])
analytics_reports = import_module('analytics.reports', ['ProductReportGenerator', 'UserReportGenerator'])
basket_reports = import_module('basket.reports', ['OpenBasketReportGenerator', 'SubmittedBasketReportGenerator'])     
offer_reports = import_module('offer.reports', ['OfferReportGenerator'])
voucher_reports = import_module('voucher.reports', ['VoucherReportGenerator'])  

class GeneratorRepository(object):
    
    generators = [order_reports.OrderReportGenerator,
                  analytics_reports.ProductReportGenerator,
                  analytics_reports.UserReportGenerator,
                  basket_reports.OpenBasketReportGenerator,
                  basket_reports.SubmittedBasketReportGenerator,
                  voucher_reports.VoucherReportGenerator, 
                  offer_reports.OfferReportGenerator]

    def get_report_generators(self):
        return self.generators
    
    def get_generator(self, code):
        for generator in self.generators:
            if generator.code == code:
                return generator
        return None
    
    



from django.db import models


class CommunicationTypeManager(models.Manager):
    
    def get_and_render(self, code, context):
        """
        Return a dictionary of rendered messages, ready for sending
        """
        try:
            commtype = self.get(code=code)
        except self.model.DoesNotExist:
            commtype = self.model(code=code)
        return commtype.get_messages(context)



from decimal import Decimal, ROUND_DOWN, ROUND_UP
import math
import datetime

from django.core import exceptions
from django.db import models
from django.utils.translation import ugettext as _
from django.core.exceptions import ValidationError
from django.conf import settings

from oscar.apps.offer.managers import ActiveOfferManager
from oscar.models.fields import PositiveDecimalField, ExtendedURLField

SITE, VOUCHER, USER, SESSION = ("Site", "Voucher", "User", "Session")


class ConditionalOffer(models.Model):
    """
    A conditional offer (eg buy 1, get 10% off)
    """
    name = models.CharField(max_length=128)
    description = models.TextField(blank=True, null=True)

    # Offers come in a few different types:
    # (a) Offers that are available to all customers on the site.  Eg a 
    #     3-for-2 offer.
    # (b) Offers that are linked to a voucher, and only become available once
    #     that voucher has been applied to the basket
    # (c) Offers that are linked to a user.  Eg, all students get 10% off.  The code
    #     to apply this offer needs to be coded
    # (d) Session offers - these are temporarily available to a user after some trigger 
    #     event.  Eg, users coming from some affiliate site get 10% off.     
    TYPE_CHOICES = (
        (SITE, "Site offer - available to all users"),
        (VOUCHER, "Voucher offer - only available after entering the appropriate voucher code"),
        (USER, "User offer - available to certain types of user"),
        (SESSION, "Session offer - temporary offer, available for a user for the duration of their session"),
    )
    offer_type = models.CharField(_("Type"), choices=TYPE_CHOICES, default=SITE, max_length=128)

    condition = models.ForeignKey('offer.Condition')
    benefit = models.ForeignKey('offer.Benefit')

    # Range of availability.  Note that if this is a voucher offer, then these
    # dates are ignored and only the dates from the voucher are used to determine 
    # availability.
    start_date = models.DateField(blank=True, null=True)
    end_date = models.DateField(blank=True, null=True)

    # Some complicated situations require offers to be applied in a set order.
    priority = models.IntegerField(default=0, help_text="The highest priority offers are applied first")

    # We track some information on usage
    total_discount = models.DecimalField(decimal_places=2, max_digits=12, default=Decimal('0.00'))
    
    date_created = models.DateTimeField(auto_now_add=True)

    objects = models.Manager()
    active = ActiveOfferManager()

    redirect_url = ExtendedURLField(_('URL redirect (optional)'), blank=True)

    # We need to track the voucher that this offer came from (if it is a voucher offer)
    _voucher = None

    class Meta:
        ordering = ['-priority']
        
    def __unicode__(self):
        return self.name    

    def clean(self):
        if self.start_date and self.end_date and self.start_date > self.end_date:
            raise exceptions.ValidationError('End date should be later than start date')
        
    def is_active(self, test_date=None):
        if not test_date:
            test_date = datetime.date.today()
        return self.start_date <= test_date and test_date < self.end_date
    
    def is_condition_satisfied(self, basket):
        return self._proxy_condition().is_satisfied(basket)
        
    def apply_benefit(self, basket):
        u"""
        Applies the benefit to the given basket and returns the discount.
        """
        if not self.is_condition_satisfied(basket):
            return Decimal('0.00')
        return self._proxy_benefit().apply(basket, self._proxy_condition())
        
    def set_voucher(self, voucher):
        self._voucher = voucher
        
    def get_voucher(self):
        return self._voucher        
        
    def _proxy_condition(self):
        u"""
        Returns the appropriate proxy model for the condition
        """
        field_dict = self.condition.__dict__
        if '_state' in field_dict:
            del field_dict['_state']
        if self.condition.type == self.condition.COUNT:
            return CountCondition(**field_dict)
        elif self.condition.type == self.condition.VALUE:
            return ValueCondition(**field_dict)
        elif self.condition.type == self.condition.COVERAGE:
            return CoverageCondition(**field_dict)
        return self.condition
    
    def _proxy_benefit(self):
        u"""
        Returns the appropriate proxy model for the condition
        """
        field_dict = self.benefit.__dict__
        if '_state' in field_dict:
            del field_dict['_state']
        if self.benefit.type == self.benefit.PERCENTAGE:
            return PercentageDiscountBenefit(**field_dict)
        elif self.benefit.type == self.benefit.FIXED:
            return AbsoluteDiscountBenefit(**field_dict)
        elif self.benefit.type == self.benefit.MULTIBUY:
            return MultibuyDiscountBenefit(**field_dict)
        elif self.benefit.type == self.benefit.FIXED_PRICE:
            return FixedPriceBenefit(**field_dict)
        return self.benefit
        

class Condition(models.Model):
    COUNT, VALUE, COVERAGE = ("Count", "Value", "Coverage")
    TYPE_CHOICES = (
        (COUNT, _("Depends on number of items in basket that are in condition range")),
        (VALUE, _("Depends on value of items in basket that are in condition range")),
        (COVERAGE, _("Needs to contain a set number of DISTINCT items from the condition range"))
    )
    range = models.ForeignKey('offer.Range')
    type = models.CharField(max_length=128, choices=TYPE_CHOICES)
    value = PositiveDecimalField(decimal_places=2, max_digits=12)

    def __unicode__(self):
        if self.type == self.COUNT:
            return u"Basket includes %d item(s) from %s" % (self.value, unicode(self.range).lower())
        elif self.type == self.COVERAGE:
            return u"Basket includes %d distinct products from %s" % (self.value, unicode(self.range).lower())
        return u"Basket includes %d value from %s" % (self.value, unicode(self.range).lower())
    
    def consume_items(self, basket, lines=None):
        return ()
    
    def is_satisfied(self, basket):
        """
        Determines whether a given basket meets this condition.  This is
        stubbed in this top-class object.  The subclassing proxies are
        responsible for implementing it correctly.
        """
        return False
    

class Benefit(models.Model):
    PERCENTAGE, FIXED, MULTIBUY, FIXED_PRICE = ("Percentage", "Absolute", "Multibuy", "Fixed price")
    TYPE_CHOICES = (
        (PERCENTAGE, _("Discount is a % of the product's value")),
        (FIXED, _("Discount is a fixed amount off the product's value")),
        (MULTIBUY, _("Discount is to give the cheapest product for free")),
        (FIXED_PRICE, _("Get the products that meet the condition for a fixed price")),
    )
    range = models.ForeignKey('offer.Range', null=True, blank=True)
    type = models.CharField(max_length=128, choices=TYPE_CHOICES)
    value = PositiveDecimalField(decimal_places=2, max_digits=12,
                                 null=True, blank=True)

    price_field = 'price_incl_tax'

    # If this is not set, then there is no upper limit on how many products 
    # can be discounted by this benefit.
    max_affected_items = models.PositiveIntegerField(blank=True, null=True, help_text="""Set this
        to prevent the discount consuming all items within the range that are in the basket.""")
    
    def __unicode__(self):
        if self.type == self.PERCENTAGE:
            desc = u"%s%% discount on %s" % (self.value, unicode(self.range).lower())
        elif self.type == self.MULTIBUY:
            desc = u"Cheapest product is free from %s" % unicode(self.range).lower()
        elif self.type == self.FIXED_PRICE:
            desc = u"The products that meet the condition are sold for %s" % self.value
        else:
            desc = u"%.2f discount on %s" % (float(self.value), unicode(self.range).lower())
        if self.max_affected_items == 1:
            desc += u" (max 1 item)"
        elif self.max_affected_items > 1:
            desc += u" (max %d items)" % self.max_affected_items
        return desc
    
    def apply(self, basket, condition=None):
        return Decimal('0.00')
    
    def clean(self):
        if self.value is None:
            if not self.type:
                raise ValidationError("Benefit requires a value")
            elif self.type != self.MULTIBUY:
                raise ValidationError("Benefits of type %s need a value" % self.type)
        elif self.value > 100 and self.type == 'Percentage':
            raise ValidationError("Percentage benefit value can't be greater than 100")
        # All benefits need a range apart from FIXED_PRICE
        if self.type and self.type != self.FIXED_PRICE and not self.range:
            raise ValidationError("Benefits of type %s need a range" % self.type)

    def round(self, amount):
        """
        Apply rounding to discount amount
        """
        if hasattr(settings, 'OSCAR_OFFER_ROUNDING_FUNCTION'):
            return settings.OSCAR_OFFER_ROUNDING_FUNCTION(amount)
        return amount.quantize(Decimal('.01'), ROUND_DOWN)

    def _effective_max_affected_items(self):
        if not self.max_affected_items:
            max_affected_items = 10000
        else:
            max_affected_items = self.max_affected_items
        return max_affected_items


class Range(models.Model):
    """
    Represents a range of products that can be used within an offer
    """
    name = models.CharField(_("Name"), max_length=128, unique=True)
    includes_all_products = models.BooleanField(default=False)
    included_products = models.ManyToManyField('catalogue.Product', related_name='includes', blank=True)
    excluded_products = models.ManyToManyField('catalogue.Product', related_name='excludes', blank=True)
    classes = models.ManyToManyField('catalogue.ProductClass', related_name='classes', blank=True)
    included_categories = models.ManyToManyField('catalogue.Category', related_name='includes', blank=True)
    
    __included_product_ids = None
    __excluded_product_ids = None
    __class_ids = None

    def __unicode__(self):
        return self.name    
        
    def contains_product(self, product):
        """
        Check whether the passed product is part of this range
        """
        # We look for shortcircuit checks first before
        # the tests that require more database queries.

        if settings.OSCAR_OFFER_BLACKLIST_PRODUCT and \
            settings.OSCAR_OFFER_BLACKLIST_PRODUCT(product):
            return False
        excluded_product_ids = self._excluded_product_ids()
        if product.id in excluded_product_ids:
            return False
        if self.includes_all_products:
            return True
        if product.product_class_id in self._class_ids():
            return True   
        included_product_ids = self._included_product_ids()
        if product.id in included_product_ids:
            return True
        test_categories = self.included_categories.all() 
        if test_categories:
            for category in product.categories.all():
                for test_category in test_categories:
                    if category == test_category or category.is_descendant_of(test_category):
                        return True
        return False
    
    def _included_product_ids(self):
        if None == self.__included_product_ids:
            self.__included_product_ids = [row['id'] for row in self.included_products.values('id')]
        return self.__included_product_ids
    
    def _excluded_product_ids(self):
        if None == self.__excluded_product_ids:
            self.__excluded_product_ids = [row['id'] for row in self.excluded_products.values('id')]
        return self.__excluded_product_ids
    
    def _class_ids(self):
        if None == self.__class_ids:
            self.__class_ids = [row['id'] for row in self.classes.values('id')]
        return self.__class_ids
        

class CountCondition(Condition):
    """
    An offer condition dependent on the NUMBER of matching items from the basket.
    """

    class Meta:
        proxy = True

    def is_satisfied(self, basket):
        """
        Determines whether a given basket meets this condition
        """
        num_matches = 0
        for line in basket.all_lines():
            if self.range.contains_product(line.product) and line.quantity_without_discount > 0:
                num_matches += line.quantity_without_discount
            if num_matches >= self.value:
                return True
        return False
    
    def consume_items(self, basket, lines=None, value=None):
        """
        Marks items within the basket lines as consumed so they
        can't be reused in other offers.
        """
        lines = lines or basket.all_lines()
        consumed_products = []
        value = self.value if value is None else value
        for line in lines:
            if self.range.contains_product(line.product):
                quantity_to_consume = min(line.quantity_without_discount,
                                          value - len(consumed_products))
                line.consume(quantity_to_consume)
                consumed_products.extend((line.product,)*int(quantity_to_consume))
            if len(consumed_products) == value:
                break
        return consumed_products
        
        
class CoverageCondition(Condition):
    """
    An offer condition dependent on the NUMBER of matching items from the basket.
    """
    class Meta:
        proxy = True

    def is_satisfied(self, basket):
        """
        Determines whether a given basket meets this condition
        """
        covered_ids = []
        for line in basket.all_lines():
            if not line.is_available_for_discount:
                continue
            product = line.product
            if self.range.contains_product(product) and product.id not in covered_ids:
                covered_ids.append(product.id)
            if len(covered_ids) >= self.value:
                return True
        return False
    
    def consume_items(self, basket, lines=None, value=None):
        """
        Marks items within the basket lines as consumed so they
        can't be reused in other offers.
        """
        lines = lines or basket.all_lines()
        consumed_products = []
        value = self.value if value is None else value
        for line in basket.all_lines():
            product = line.product
            if (line.is_available_for_discount and self.range.contains_product(product)
                and product not in consumed_products):
                line.consume(1)
                consumed_products.append(line.product)
            if len(consumed_products) >= value:
                break
        return consumed_products
    
    def get_value_of_satisfying_items(self, basket):
        covered_ids = []
        value = Decimal('0.00')
        for line in basket.all_lines():
            if self.range.contains_product(line.product) and line.product.id not in covered_ids:
                covered_ids.append(line.product.id)
                value += line.unit_price_incl_tax
            if len(covered_ids) >= self.value:
                return value
        return value
        
        
class ValueCondition(Condition):
    """
    An offer condition dependent on the VALUE of matching items from the basket.
    """
    price_field = 'price_incl_tax'

    class Meta:
        proxy = True

    def is_satisfied(self, basket):
        """Determines whether a given basket meets this condition"""
        value_of_matches = Decimal('0.00')
        for line in basket.all_lines():
            product = line.product
            if self.range.contains_product(product) and product.has_stockrecord and line.quantity_without_discount > 0:
                price = getattr(product.stockrecord, self.price_field)
                value_of_matches += price * int(line.quantity_without_discount)
            if value_of_matches >= self.value:
                return True
        return False
    
    def consume_items(self, basket, lines=None, value=None):
        """
        Marks items within the basket lines as consumed so they
        can't be reused in other offers.
        
        We allow lines to be passed in as sometimes we want them sorted
        in a specific order.
        """
        value_of_matches = Decimal('0.00')
        lines = lines or basket.all_lines()
        consumed_products = []
        value = self.value if value is None else value
        for line in basket.all_lines():
            product = line.product
            if self.range.contains_product(product) and line.product.has_stockrecord:
                price = getattr(product.stockrecord, self.price_field)
                if not price:
                    continue
                quantity_to_consume = min(line.quantity_without_discount,
                                          int(((value - value_of_matches)/price).quantize(Decimal(1),
                                                                                              ROUND_UP)))
                value_of_matches += price * quantity_to_consume
                line.consume(quantity_to_consume)
                consumed_products.extend((line.product,)*int(quantity_to_consume))
            if value_of_matches >= value:
                break
        return consumed_products

# ========
# Benefits
# ========

class PercentageDiscountBenefit(Benefit):
    """
    An offer benefit that gives a percentage discount
    """

    class Meta:
        proxy = True

    def apply(self, basket, condition=None):
        discount = Decimal('0.00')
        affected_items = 0
        max_affected_items = self._effective_max_affected_items()
        
        for line in basket.all_lines():
            if affected_items >= max_affected_items:
                break
            product = line.product
            if self.range.contains_product(product) and product.has_stockrecord:
                price = getattr(product.stockrecord, self.price_field)
                quantity = min(line.quantity_without_discount, 
                               max_affected_items - affected_items)
                line_discount = self.round(self.value/100 * price * int(quantity))
                line.discount(line_discount, quantity)
                affected_items += quantity
                discount += line_discount
                
        if discount > 0 and condition:
            condition.consume_items(basket)
        return discount


class AbsoluteDiscountBenefit(Benefit):
    """
    An offer benefit that gives an absolute discount
    """

    class Meta:
        proxy = True

    def apply(self, basket, condition=None):
        discount = Decimal('0.00')
        affected_items = 0
        max_affected_items = self._effective_max_affected_items()
        
        for line in basket.all_lines():
            if affected_items >= max_affected_items:
                break
            product = line.product
            if self.range.contains_product(product) and product.has_stockrecord:
                price = getattr(product.stockrecord, self.price_field)
                if not price:
                    # Avoid zero price products
                    continue
                remaining_discount = self.value - discount
                quantity_affected = int(min(line.quantity_without_discount, 
                                        max_affected_items - affected_items,
                                        math.ceil(remaining_discount / price)))
                
                # Update line with discounts
                line_discount = self.round(min(remaining_discount, quantity_affected * price))
                line.discount(line_discount, quantity_affected)
                
                # Update loop vars
                affected_items += quantity_affected
                remaining_discount -= line_discount
                discount += line_discount
        if discount > 0 and condition:
            condition.consume_items(basket)
            
        return discount


class FixedPriceBenefit(Benefit):
    """
    An offer benefit that gives the items in the condition for a 
    fixed price.  This is useful for "bundle" offers.
    
    Note that we ignore the benefit range here and only give a fixed price
    for the products in the condition range.
    
    The condition should be a count condition
    """
    class Meta:
        proxy = True

    def apply(self, basket, condition=None):
        num_covered = 0
        num_permitted = int(condition.value)
        covered_lines = []
        product_total = Decimal('0.00')
        for line in basket.all_lines():
            product = line.product
            if condition.range.contains_product(product) and line.quantity_without_discount > 0:
                # Line is available - determine quantity to consume and 
                # record the total of the consumed products
                if isinstance(condition, CoverageCondition):
                    quantity = 1
                else:
                    quantity = min(line.quantity_without_discount, num_permitted - num_covered)
                num_covered += quantity
                product_total += quantity*line.unit_price_incl_tax
                covered_lines.append((line, quantity))
            if num_covered >= num_permitted:
                break
        discount = max(product_total - self.value, Decimal('0.00'))
        
        if not discount:
            return discount
        
        # Apply discount weighted by original value of line
        discount_applied = Decimal('0.00')
        last_line = covered_lines[-1][0]
        for line, quantity in covered_lines:
            if line == last_line:
                # If last line, we just take the difference to ensure that
                # a rounding doesn't lead to an off-by-one error
                line_discount = discount - discount_applied
            else:
                line_discount = self.round(discount * (line.unit_price_incl_tax * quantity) / product_total)
            line.discount(line_discount, quantity)
            discount_applied += line_discount
        return discount 
        
        # Apply discount weighted by original value of line
        for line, quantity in covered_lines:
            line_discount = self.round(discount * (line.unit_price_incl_tax * quantity) / product_total)  
            line.discount(line_discount.quantize(Decimal('.01')), quantity)
        return discount 


class MultibuyDiscountBenefit(Benefit):
    class Meta:
        proxy = True

    def apply(self, basket, condition=None):
        benefit_lines = [line for line in basket.all_lines() if (self.range.contains_product(line.product) and
                                                                 line.quantity_without_discount > 0 and
                                                                 line.product.has_stockrecord)]
        if not benefit_lines:
            return self.round(Decimal('0.00'))

        # Determine cheapest line to give for free
        line_price_getter = lambda line: getattr(line.product.stockrecord,
                                                 self.price_field)
        free_line = min(benefit_lines, key=line_price_getter)
        discount = line_price_getter(free_line)

        if condition:
            compare = lambda l1, l2: cmp(line_price_getter(l2),
                                         line_price_getter(l1))
            lines_with_price = [line for line in basket.all_lines() if line.product.has_stockrecord]
            sorted_lines = sorted(lines_with_price, compare)
            free_line.discount(discount, 1)
            if condition.range.contains_product(line.product):
                condition.consume_items(basket, lines=sorted_lines,
                                        value=condition.value-1)
            else:
                condition.consume_items(basket, lines=sorted_lines)
        else:
            free_line.discount(discount, 0)
        return self.round(discount)


# We need to import receivers at the bottom of this script
from oscar.apps.offer.receivers import receive_basket_voucher_change
